# THANKS to https://github.com/AlexxIT/XiaomiGateway3 for the HASS implementation
# which I used as a base for the FHEM implementation

import base64
import json
import asyncio
import socket
import re
import functools
from telnetlib import Telnet
import time
from asyncio_mqtt import Client, MqttError

from .miio_fix import Device
from .unqlite import Unqlite

from .utils import GLOBAL_PROP
from . import utils

from .. import fhem
from .. import utils as fpyutils

RE_NWK_KEY = re.compile(r'lumi send-nwk-key (0x.+?) {(.+?)}')

class xiaomi_gateway3:

  def __init__(self, logger):
    self.logger = logger
    return

  # FHEM FUNCTION
  async def Define(self, hash, args, argsh):
    self.hash = hash

    if len(args) < 5:
      return "Usage: define devname PythonModule xiaomi_gateway3 <IP> <token>"

    if await fhem.AttrVal(self.hash['NAME'], "icon", "") == "":
      await fhem.CommandAttr(self.hash, self.hash["NAME"] + " icon mqtt")
    await fhem.readingsSingleUpdateIfChanged(hash, "state", "disconnected", 1)
    
    hash["HOST"] = args[3]
    hash["TOKEN"] = args[4]

    self.host = args[3]
    self.token = args[4]

    asyncio.create_task(self.connect_gw())
    
    return ""

  # FHEM FUNCTION
  async def Undefine(self, hash):
    return

  # FHEM FUNCTION
  async def Set(self, hash, args, argsh):
    return ""

  def register_device(self, did, upd_listener):
    self.gw.register_device(did, upd_listener)
  
  def get_device(self, did):
    return self.gw.get_device(did)

  async def connect_gw(self):
    self.gw = Gateway(self.logger, self.hash, self.host, self.token)
    # connect to gateway
    await self.gw.connect()
    # create task which handles MQTT messages
    asyncio.create_task(self.gw.connect_mqtt())

class Gateway:

  def __init__(self, logger, hash, host, token):
    self.hash = hash
    self.logger = logger
    self.host = host
    self.token = token
    self.devices = {}
    self.miio = Device(host, token)
    self.child_devices = {}
    self.connected = False
    self.pair_model = None
    self.pair_payload = None

  def register_device(self, did, upd_listener):
    if did not in self.child_devices:
      self.child_devices[did] = []
    self.child_devices[did].append(upd_listener)
    # report initial states
    asyncio.create_task(self.report(did))

  def get_device(self, did):
    if did in self.devices:
      return self.devices[did]
    return None

  async def connect(self):
    # first connect
    await fpyutils.run_blocking(functools.partial(self.thread_blocking_connect))
    await fhem.readingsSingleUpdateIfChanged(self.hash, "state", "connected", 1)
    await self.create_devices()
    await self.report_all()
    # start check_connection task to reconnect on gw reboot
    asyncio.create_task(self.check_connection())

  async def check_connection(self):
    while True:
      if self.connected is False:
        await fpyutils.run_blocking(functools.partial(self.thread_blocking_connect))
        if self.connected:
          await self.create_devices()
          await self.report_all()
      # sleep 30s
      await asyncio.sleep(30)

  def thread_blocking_connect(self, keepconnection=0):
    while self.connected is False:
      if self._miio_connect():
          self.update_devices()
      else:
          time.sleep(30)
    # connected

  def update_devices(self):
    devices = self._get_devices_v3()
    if devices is None:
        self._enable_telnet()
        self._enable_mqtt()
    else:
        self.setup_devices(devices)

  def setup_devices(self, devices: list):
    """Add devices to hass."""
    for device in devices:
        desc = utils.get_device(device['model'])
        if not desc:
            self.logger.debug(f"Unsupported model: {device}")
            continue

        self.logger.debug(f"{self.host} | Setup device {device['model']}")

        device.update(desc)

        # update params from config
        default_config = self.devices.get(device['mac'])
        if default_config:
            device.update(default_config)

        self.devices[device['did']] = device

  async def create_devices(self):
    for did in self.devices:
      if not await fhem.checkIfDeviceExists(self.hash, "PYTHONTYPE", "xiaomi_gateway3_device", "DID", did):
        devname = "".join(filter(str.isalnum, self.devices[did]['model'])) + "_" + self.devices[did]['sid']
        await fhem.CommandDefine(self.hash, devname + " PythonModule xiaomi_gateway3_device " + self.hash["NAME"] + " " + did)
        # avoid fhem commands overload
        await asyncio.sleep(3)
      # avoid fhem commands overload
      await asyncio.sleep(1)

  async def report(self, did):
    if did in self.devices:
      device = self.devices[did]
      if "init" in device:
        if did in self.child_devices:
          for listener in self.child_devices[did]:
            await listener.update(device['init'])

  async def report_all(self):
    for did in self.devices:
      await self.report(did)

  def _miio_connect(self) -> bool:
    try:
        self.miio.send_handshake()
        return True
    except:
        self.logger.debug(f"{self.host} | Can't send handshake")
        return False

  def _enable_telnet(self):
      self.logger.debug(f"{self.host} | Try enable telnet")
      try:
          resp = self.miio.send("enable_telnet_service")
          return resp[0] == 'ok'
      except Exception as e:
          self.logger.exception(f"Can't enable telnet: {e}")
          return False

  def _enable_mqtt(self):
    self.logger.debug(f"{self.host} | Try run public MQTT")
    try:
        telnet = Telnet(self.host)
        telnet.read_until(b"login: ")
        telnet.write(b"admin\r\n")
        telnet.read_until(b"\r\n# ")  # skip greeting

        # enable public mqtt
        telnet.write(b"killall mosquitto\r\n")
        telnet.read_until(b"\r\n")  # skip command
        time.sleep(.5)  # it's important to wait
        telnet.write(b"mosquitto -d\r\n")
        telnet.read_until(b"\r\n")  # skip command
        time.sleep(.5)  # it's important to wait

        # fix CPU 90% full time bug
        telnet.write(b"killall zigbee_gw\r\n")
        telnet.read_until(b"\r\n")  # skip command
        time.sleep(.5)  # it's important to wait

        telnet.close()
        return True
    except Exception as e:
        self.logger.debug(f"Can't run MQTT: {e}")
        return False

  async def connect_mqtt(self):
    try:
      async with Client(self.host) as client:
        self.mqtt = client
        async with client.filtered_messages("#") as messages:
          await client.subscribe("#")
          async for message in messages:
            self.logger.debug(f"Message from MQTT: {message.topic}: {message.payload.decode()}")
            try:
              if message.topic == 'zigbee/send':
                payload = json.loads(message.payload.decode())
                # convert to miio msg
                msg = self.process_message(payload)
                if msg['did'] in self.child_devices:
                  for listener in self.child_devices[msg['did']]:
                    await listener.update(msg['data'])
              elif message.topic.endswith('/heartbeat'):
                for listener in self.child_devices['lumi.0']:
                  await listener.update(json.loads(message.payload.decode()))
              elif self.pair_model and message.topic.endswith('/commands'):
                self.process_pair(message.payload)
            except:
                self.logger.exception("Failed to handle MQTT message")
    except:
      self.connected = False
      self.logger.exception("Failed to connect to MQTT server: " + self.host)

  def process_pair(self, raw: bytes):
    # get shortID and eui64 of paired device
    if b'lumi send-nwk-key' in raw:
        # create model response
        payload = f"0x18010105000042{len(self.pair_model):02x}" \
                  f"{self.pair_model.encode().hex()}"
        m = RE_NWK_KEY.search(raw.decode())
        self.pair_payload = json.dumps({
            'sourceAddress': m[1],
            'eui64': '0x' + m[2],
            'profileId': '0x0104',
            'clusterId': '0x0000',
            'sourceEndpoint': '0x01',
            'destinationEndpoint': '0x01',
            'APSCounter': '0x01',
            'APSPlayload': payload
        }, separators=(',', ':'))

    # send model response "from device"
    elif b'zdo active ' in raw:
        mac = '' #self.device['mac'][2:].upper()
        self.mqtt.publish(f"gw/{mac}/MessageReceived", self.pair_payload)

  def _get_devices_v3(self):
        """Load device list via Telnet."""
        self.logger.debug(f"{self.host} | Read devices")
        try:
            telnet = Telnet(self.host, timeout=5)
            telnet.read_until(b"login: ")
            telnet.write(b"admin\r\n")
            telnet.read_until(b'\r\n# ')  # skip greeting

            # https://github.com/AlexxIT/XiaomiGateway3/issues/14
            # fw 1.4.6_0012 and below have one zigbee_gw.db file
            # fw 1.4.6_0030 have many json files in this folder
            telnet.write(b"cat /data/zigbee_gw/* | base64\r\n")
            telnet.read_until(b'\r\n')  # skip command
            raw = telnet.read_until(b'# ')
            raw = base64.b64decode(raw)
            if raw.startswith(b'unqlite'):
                db = Unqlite(raw)
                data = db.read_all()
            else:
                raw = re.sub(br'}\s+{', b',', raw)
                data = json.loads(raw)

            devices = []

            # data = {} or data = {'dev_list': 'null'}
            dev_list = json.loads(data.get('dev_list', 'null')) or []
            self.logger.debug(f"{self.host} | Load {len(dev_list)} zigbee devices")

            for did in dev_list:
                model = data[did + '.model']
                desc = utils.get_device(model)

                # skip unknown model
                if desc is None:
                    self.logger.debug(f"Unsupported model: {model}")
                    continue

                retain = json.loads(data[did + '.prop'])['props']

                params = retain
                # params = {
                #     p[2]: retain.get(p[1])
                #     for p in desc['params']
                #     if p[1] is not None
                # }

                # fix some param values
                for k, v in list(params.items()):
                    if k in ('temperature', 'humidity', 'pressure'):
                        params[k] = v / 100.0
                    elif k in ('voltage'):
                        params[k] = v / 1000.0
                    elif k in ('battery') and v:
                        params['batteryPercentage'] = params['battery']
                        params['battery'] = "low" if params['batteryPercentage'] < 25 else "ok"
                    elif k in ('status'):
                        params['state'] = params['status']
                        del params['status']

                self.logger.debug(f"{self.host} | {model} retain: {params}")

                device = {
                    'did': did,
                    'sid': did.split(".")[1],
                    'mac': '0x' + data[did + '.mac'],
                    'model': data[did + '.model'],
                    'type': 'zigbee',
                    'zb_ver': data[did + '.version'],
                    'init': params
                }
                devices.append(device)

            telnet.write(b"cat /data/zigbee/coordinator.info\r\n")
            telnet.read_until(b'\r\n')  # skip command
            raw = telnet.read_until(b'# ')

            device = self.miio.info()
            devices.append({
                'did': 'lumi.0',
                'sid': '0',
                'mac': device.mac_address,  # wifi mac!!!
                'model': device.model
            })

            self.connected = True

            return devices

        except (ConnectionRefusedError, socket.timeout):
            return None

        except Exception as e:
            self.logger.debug(f"Can't read devices: {e}")
            return None

  def process_message(self, data: dict):
        cmd = data['cmd']
        if data['cmd'] == 'heartbeat':
            # don't know if only one item
            assert len(data['params']) == 1, data

            data = data['params'][0]
            pkey = 'res_list'
        elif data['cmd'] == 'report':
            pkey = 'params' if 'params' in data else 'mi_spec'
        elif data['cmd'] == 'write_rsp':
            pkey = 'results'
        else:
            raise Exception(f"Unsupported cmd: {data}")

        did = data['did']

        device = self.devices[did]
        payload = {}

        # convert codes to names
        for param in data[pkey]:
            if param.get('error_code', 0) != 0:
                continue

            prop = param['res_name'] if 'res_name' in param else \
                f"{param['siid']}.{param['piid']}"

            if prop in GLOBAL_PROP:
                prop = GLOBAL_PROP[prop]
            else:
                prop = next((p[2] for p in device['params']
                             if p[0] == prop), prop)

            if prop in ('temperature', 'humidity', 'pressure'):
                payload[prop] = param['value'] / 100.0
            elif prop in ('voltage'):
                payload[prop] = param['value'] / 1000.0
            elif prop in ('battery'):
                # xiaomi light sensor
                payload['batteryPercentage'] = param['value']
                payload['battery'] = "low" if payload['batteryPercentage'] < 25 else "ok"
            elif prop in ('status'):
                payload['state'] = param['value']
            elif prop == 'contact':
                payload['state'] = "open" if param['value'] == 1 else "close"
            elif prop == 'angle':
                # xiaomi cube 100 points = 360 degrees
                payload[prop] = param['value'] * 4
            elif prop == 'duration':
                # xiaomi cube
                payload[prop] = param['value'] / 1000.0
            else:
                payload[prop] = param['value']

        self.logger.debug(f"{self.host} | {device['did']} {device['model']} <= "
                      f"{payload}")

        if did in self.devices:
          self.devices[did]['init'] = payload
        else:
          # TODO create device
          pass

        return {
          "did": did,
          "sid": device['sid'],
          "model": device['model'],
          "cmd": cmd,
          "data": payload
        }




import asyncio
import functools

from micloud import MiCloud

from .. import utils
from .. import fhem

class xiaomi_tokens:

    def __init__(self, logger):
        self.logger = logger
        self._username = None
        self._password = None
        self._country = "de"
        return

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        self.hash = hash
        await fhem.readingsSingleUpdateIfChanged(hash, "state", "active", 1)
        return ""

    # FHEM FUNCTION
    async def Undefine(self, hash):
        return

    # FHEM FUNCTION
    async def Set(self, hash, args, argsh):
        set_list_conf = {
           "username": { "args": ["username"] },
           "password": { "args": ["password"] },
           "country": { "args": ["country"], "options": "de,cn,sg" },
           "get_tokens": {}
        }
        return await utils.handle_set(set_list_conf, self, hash, args, argsh)

    async def set_username(self, hash, params):
        self._username = params['username']
        return ""

    async def set_password(self, hash, params):
        self._password = params['password']
        return ""

    async def set_country(self, hash, params):
        self._country = params['country']
        return ""

    async def set_get_tokens(self, hash):
        if self._username and self._password:
            asyncio.create_task(self.obtain_tokens())
        else:
            return "Please set username & password first!"

    async def obtain_tokens(self):
        await utils.run_blocking(functools.partial(self.thread_get_tokens))
        await fhem.readingsBeginUpdate(self.hash)
        for dev in self._device_list:
            await fhem.readingsBulkUpdateIfChanged(self.hash, dev['did'] + "_name", dev['name'])
            await fhem.readingsBulkUpdateIfChanged(self.hash, dev['did'] + "_token", dev['token'])
            await fhem.readingsBulkUpdateIfChanged(self.hash, dev['did'] + "_model", dev['model'])
        await fhem.readingsEndUpdate(self.hash, 1)
    
    def thread_get_tokens(self):
        mc = MiCloud(self._username, self._password)
        mc.login()
        self._device_list = mc.get_devices(country=self._country)



from ..generic import FhemModule
import asyncio
import functools
import logging

from .. import fhem
from .. import utils as fpyutils

from .nespresso import NespressoDetect

class nespresso_ble(FhemModule):

    def __init__(self, logger):
        super().__init__(logger)
        self.nespressodetect = None
        self.task = None
        self.auth = None
        logging.getLogger("pygatt.backends.gatttool.gatttool").setLevel(logging.ERROR)
        #logging.getLogger("nespresso_ble").setLevel(logging.DEBUG)
        self.set_conf_list = {
          "authkey": { "args": ["authkey"] },
          "brew": {"args": ["coffee_type", "temperature"], "params": {"temperature": {"default":"high", "optional":True}, "coffee_type": {"default":"lungo", "optional":True}},
            "help": "1. parameter: ristretto, espresso, lung, hotwater, americano<br>2. parameter: Temperature with values low, mid, high"},
          "easybrew": {"args": ["coffee_type"], "options": "ristretto,espresso,lungo,hotwater,americano"},
          "recipe": { "help": "Not yet supported" },
          "updateStatus": {}
        }
        self.set_set_config(self.set_conf_list)
        return

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
      await super().Define(hash, args, argsh)
      await fhem.readingsBeginUpdate(hash)
      await fhem.readingsBulkUpdateIfChanged(hash, "state", "offline")
      await fhem.readingsEndUpdate(hash, 1)
      self.hash = hash
      if len(args) < 4:
        return "Usage: define devicename PythonModule nespresso_ble <MAC> [<AUTHKEY>]"
      self.mac = args[3]
      hash["MAC"] = args[3]

      # check if there is already an authkey
      if len(args) > 4:
        self.auth = args[4]
      else:
        self.auth = await fhem.ReadingsVal(self.hash['NAME'], "authkey", "")

      if self.auth != "":
        self.auth = args[4]
        self.nespressodetect = NespressoDetect(self.auth, self.mac)
        self.nespressodetect.set_keep_connected(True)
        self.task = self.create_async_task(self.update_status_task())
      return ""

    async def update_status_task(self):
      while True:
        if self.nespressodetect:
          await self.update_status()
        await asyncio.sleep(300)

    # FHEM FUNCTION
    async def Set(self, hash, args, argsh):
      if self.auth and "authkey" in self.set_conf_list:
        del self.set_conf_list['authkey']
        self.set_set_config(self.set_conf_list)
      return await super().Set(hash, args, argsh)

    async def set_authkey(self, hash, params):
      self.auth = params["authkey"]
      self.set_set_config(self.set_conf_list)
      await fhem.readingsSingleUpdateIfChanged(self.hash, "authkey", self.auth, 1)
      if self.task:
        self.task.cancel()
      self.nespressodetect = NespressoDetect(self.auth, self.mac)
      self.nespressodetect.set_keep_connected(True)
      self.task = self.create_async_task(self.update_status_task())

    async def set_easybrew(self, hash, params):
      params['temperature'] = "medium"
      await self.set_brew(hash, params)

    async def set_brew(self, hash, params):
      try:
        coffee_type = params["coffee_type"]
        temp = params["temperature"]
        fpyutils.run_blocking_task(functools.partial(self.nespressodetect.make_coffee,self.mac, temp, coffee_type))
      except:
        await fhem.readingsSingleUpdateIfChanged(self.hash, "state", "offline", 1)

    async def set_updateStatus(self, hash):
      self.create_async_task(self.update_status())

    async def update_status(self):
      await fpyutils.run_blocking(functools.partial(self.blocking_update_status))

      if self.device_info:
        for mac, dev in self.device_info.items():
          await fhem.readingsSingleUpdateIfChanged(self.hash, "manufacturer", dev.manufacturer, 1)
          await fhem.readingsSingleUpdateIfChanged(self.hash, "serial_nr", dev.serial_nr, 1)
          await fhem.readingsSingleUpdateIfChanged(self.hash, "model_nr", dev.model_nr, 1)
          await fhem.readingsSingleUpdateIfChanged(self.hash, "device_name", dev.device_name, 1)
          if dev.device_name:
            await fhem.readingsSingleUpdateIfChanged(self.hash, "state", "online", 1)
          else:
            await fhem.readingsSingleUpdateIfChanged(self.hash, "state", "offline", 1)
      else:
        await fhem.readingsSingleUpdateIfChanged(self.hash, "state", "offline", 1)

      if self.sensors_data:
        for mac, data in self.sensors_data.items():
          for name, val in data.items():
            await fhem.readingsSingleUpdateIfChanged(self.hash, name, val, 1)

    def blocking_update_status(self):
      self.logger.debug("nespresso_ble updatestatus")
      try:
        self.device_info = self.nespressodetect.get_info()
        self.nespressodetect.get_sensors()
        self.sensors_data = self.nespressodetect.get_sensor_data()

      except:
        self.logger.exception("Failed to update status")
        self.sensors_data = None
        self.device_info = None



import asyncio
import logging
import traceback
import concurrent.futures
from zeroconf import ServiceBrowser, Zeroconf
import threading

from .. import fhem

class discover_mdns:

    def __init__(self, logger):
        self.logger = logger
        self.loop = asyncio.get_event_loop()
        self.zeroconf = None
        self.hash = None
        self.browser = None

    # zeroconf callback
    def update_service(self, zeroconf, type, name):
        info = zeroconf.get_service_info(type, name)
        self.logger.debug("Service %s updated, service info %s" % (name,info))

    # zeroconf callback
    def remove_service(self, zeroconf, type, name):
        self.logger.debug("Service %s removed" % (name))

    # zeroconf callback
    def add_service(self, zeroconf, type, name):
        info = zeroconf.get_service_info(type, name)
        self.logger.debug("Service %s added, service info: %s" % (name, info))
        res = asyncio.run_coroutine_threadsafe(self.foundDevice(name, info), self.loop)
        res.result()

    async def foundDevice(self, name, info):
        try:
            def get_value(key):
                """Retrieve value and decode to UTF-8."""
                value = info.properties.get(key.encode("utf-8"))

                if value is None or isinstance(value, str):
                    return value
                return value.decode("utf-8")

            if (info.type == "_googlecast._tcp.local."):
                # check if device exists already, if not commanddefine
                if not (await fhem.checkIfDeviceExists(self.hash, "PYTHONTYPE", "googlecast", "CASTNAME", get_value('fn'))):
                    self.logger.debug("create device: " + get_value('fn'))
                    await fhem.CommandDefine(self.hash, get_value('md').replace(" ", "_") + "_" + get_value('fn').replace(" ", "_") +  " PythonModule googlecast '" + get_value('fn') + "'")
                else:
                    self.logger.debug("device " + get_value('fn') + " exists already, do not create")
            elif (info.type == "_soundtouch._tcp.local."):
                if not (await fhem.checkIfDeviceExists(self.hash, "TYPE", "BOSEST", "DEVICEID", "0")):
                    self.logger.debug("create bosest")
                    await fhem.CommandDefine(self.hash, "bosesystem BOSEST")
                else:
                    self.logger.debug("device BOSEST exists already, do not create")
            else:
                return

            # wait for the devices to initialize
            await asyncio.sleep(10)
        except Exception as err:
            self.logger.error(traceback.print_exc())
    
    async def runZeroconfScan(self):
        # await here to finish define before zeroconf object is created
        await asyncio.sleep(1)
        self.zeroconf = Zeroconf()
        listener = self
        services = ["_googlecast._tcp.local.", "_soundtouch._tcp.local."]
        self.browser = ServiceBrowser(self.zeroconf, services, listener)

    # FHEM
    async def Define(self, hash, args, argsh):
        self.hash = hash
        await fhem.readingsSingleUpdate(self.hash, "state", "active", 1)
        self.loop.create_task(self.runZeroconfScan())
        
        if await fhem.AttrVal(self.hash['NAME'], "icon", "") == "":
            await fhem.CommandAttr(self.hash, self.hash["NAME"] + " icon rc_SEARCH")

        return ""

    # FHEM
    async def Set(self, hash, args, argsh):
        return ""

    # FHEM
    async def Undefine(self, hash):
        if self.browser:
            self.browser.cancel()
        self.zeroconf.close()



""" Contains construct adapters and structures. """
from construct import (Struct, Adapter, Int8ub, Enum, FlagsEnum, Const,
                       GreedyRange, IfThenElse, Bytes, Optional)
from datetime import datetime, time, timedelta


PROP_ID_RETURN = 1
PROP_INFO_RETURN = 2
PROP_SCHEDULE_SET = 0x10
PROP_SCHEDULE_RETURN = 0x21

NAME_TO_DAY = {"sat": 0, "sun": 1, "mon": 2, "tue": 3, "wed": 4, "thu": 5, "fri": 6}
NAME_TO_CMD = {"write": PROP_SCHEDULE_SET, "response": PROP_SCHEDULE_RETURN}
HOUR_24_PLACEHOLDER = 1234


class TimeAdapter(Adapter):
    """ Adapter to encode and decode schedule times. """
    def _decode(self, obj, ctx, path):
        h, m = divmod(obj * 10, 60)
        if h == 24:  # HACK, can we do better?
            return HOUR_24_PLACEHOLDER
        return time(hour=h, minute=m)

    def _encode(self, obj, ctx, path):
        # TODO: encode h == 24 hack
        if obj == HOUR_24_PLACEHOLDER:
            return int(24 * 60 / 10)
        encoded = int((obj.hour * 60 + obj.minute) / 10)
        return encoded

class TempAdapter(Adapter):
    """ Adapter to encode and decode temperature. """
    def _decode(self, obj, ctx, path):
        return float(obj / 2.0)

    def _encode(self, obj, ctx, path):
        return int(obj * 2.0)


class WindowOpenTimeAdapter(Adapter):
    """ Adapter to encode and decode window open times (5 min increments). """
    def _decode(self, obj, context, path):
        return timedelta(minutes=float(obj * 5.0))

    def _encode(self, obj, context, path):
        if isinstance(obj, timedelta):
            obj = obj.seconds
        if 0 <= obj <= 3600.0:
            return int(obj / 300.0)
        raise ValueError("Window open time must be between 0 and 60 minutes "
                         "in intervals of 5 minutes.")


class TempOffsetAdapter(Adapter):
    """ Adapter to encode and decode the temperature offset. """
    def _decode(self, obj, context, path):
        return float((obj - 7) / 2.0)

    def _encode(self, obj, context, path):
        if -3.5 <= obj <= 3.5:
            return int(obj * 2.0) + 7
        raise ValueError("Temperature offset must be between -3.5 and 3.5 (in "
                         "intervals of 0.5).")


ModeFlags = "ModeFlags" / FlagsEnum(Int8ub,
                                    AUTO=0x00, # always True, doesnt affect building
                                    MANUAL=0x01,
                                    AWAY=0x02,
                                    BOOST=0x04,
                                    DST=0x08,
                                    WINDOW=0x10,
                                    LOCKED=0x20,
                                    UNKNOWN=0x40,
                                    LOW_BATTERY=0x80)


class AwayDataAdapter(Adapter):
    """ Adapter to encode and decode away data. """
    def _decode(self, obj, ctx, path):
        (day, year, hour_min, month) = obj
        year += 2000

        min = 0
        if hour_min & 0x01:
            min = 30
        hour = int(hour_min / 2)

        return datetime(year=year, month=month, day=day, hour=hour, minute=min)

    def _encode(self, obj, ctx, path):
        if obj.year < 2000 or obj.year > 2099:
            raise Exception("Invalid year, possible [2000,2099]")
        year = obj.year - 2000
        hour = obj.hour * 2
        if obj.minute:  # we encode all minute values to h:30
            hour |= 0x01
        return (obj.day, year, hour, obj.month)


class DeviceSerialAdapter(Adapter):
    """ Adapter to decode the device serial number. """
    def _decode(self, obj, context, path):
        return bytearray(n - 0x30
                         for n in obj).decode()


Status = "Status" / Struct(
    "cmd" / Const(PROP_INFO_RETURN, Int8ub),
    Const(0x01, Int8ub),
    "mode" / ModeFlags,
    "valve" / Int8ub,
    Const(0x04, Int8ub),
    "target_temp" / TempAdapter(Int8ub),
    "away" / IfThenElse(lambda ctx: ctx.mode.AWAY,
                        AwayDataAdapter(Bytes(4)),
                        Optional(Bytes(4))),
    "presets" / Optional(Struct(
        "window_open_temp" / TempAdapter(Int8ub),
        "window_open_time" / WindowOpenTimeAdapter(Int8ub),
        "comfort_temp" / TempAdapter(Int8ub),
        "eco_temp" / TempAdapter(Int8ub),
        "offset" / TempOffsetAdapter(Int8ub),
    ))
)

Schedule = "Schedule" / Struct(
    "cmd" / Enum(Int8ub, **NAME_TO_CMD),
    "day" / Enum(Int8ub, **NAME_TO_DAY),
    "base_temp" / TempAdapter(Int8ub),
    "next_change_at" / TimeAdapter(Int8ub),
    "hours" / GreedyRange(Struct(
        "target_temp" / TempAdapter(Int8ub),
        "next_change_at" / TimeAdapter(Int8ub),
    )),
)

DeviceId = "DeviceId" / Struct(
    "cmd" / Const(PROP_ID_RETURN, Int8ub),
    "version" / Int8ub,
    Int8ub,
    Int8ub,
    "serial" / DeviceSerialAdapter(Bytes(10)),
    Int8ub,
)



import asyncio
import aiohttp

from .. import fhem
from .. import utils

BASE_URL = "http://www.wienerlinien.at/ogd_realtime/monitor?rbl={}"

DEPARTURES = {
    "first": {"key": 0, "name": "{} first departure"},
    "next": {"key": 1, "name": "{} next departure"},
}

class wienerlinien:

    def __init__(self, logger):
        self.logger = logger
        self.firstnext = "first"
        self._updateloop = None
        self._last_data = None
        return

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        self.hash = hash
        self._stopid = args[3]
        self.api = WienerlinienAPI(self._stopid)
        self._updateloop = asyncio.create_task(self.update_loop())
        # delete all readings on define
        asyncio.create_task(fhem.CommandDeleteReading(hash, hash['NAME'] + " .*"))

    # FHEM FUNCTION
    async def Undefine(self, hash):
        if self._updateloop:
            self._updateloop.cancel()
        return

    # FHEM FUNCTION
    async def Set(self, hash, args, argsh):
        set_list_conf = {
           "update": {}
        }
        return await utils.handle_set(set_list_conf, self, hash, args, argsh)

    async def set_update(self, hash):
        asyncio.create_task(self.update())
        return ""

    async def update_loop(self):
        while True:
            await self.update()
            await asyncio.sleep(30)

    async def update(self):
        try:
            data = await self.api.get_json()
            self.logger.debug(data)
            if data is None:
                return
            message = data.get("message", {})
            data = data.get("data", {})
        except:
            self.logger.debug("Could not get new state")
            return

        if data is None:
            return
        try:
            if len(data['monitors']) == 0:
                flat_data = {}
                flat_data_location = {}
                state_text = "no departures"
            else:
                flat_data = utils.flatten_json(data['monitors'][0]['lines'][0])
                flat_data_location = utils.flatten_json(data['monitors'][0]['locationStop'])
                state_text = flat_data['towards'] + ": " + str(flat_data['departures_departure_0_departureTime_countdown'])

            if self._last_data:
                del_readings = set(self._last_data) - set(flat_data)
            else:
                del_readings = {}

            await fhem.readingsBeginUpdate(self.hash)
            for msg in message:
                await fhem.readingsBulkUpdateIfChanged(self.hash, "msg_" + msg, message[msg])
            for data_name in flat_data:
                await fhem.readingsBulkUpdateIfChanged(self.hash, "line_" + data_name, flat_data[data_name])
            for data_name in flat_data_location:
                await fhem.readingsBulkUpdateIfChanged(self.hash, "loc_" + data_name, flat_data_location[data_name])
            
            if 'trafficjam' in flat_data and flat_data['trafficjam'] == 1:
                state_text += " (traffic jam)"
            await fhem.readingsBulkUpdateIfChanged(self.hash, "state", state_text)
            await fhem.readingsEndUpdate(self.hash, 1)

            # delete old readings which were not updated
            for del_reading in del_readings:
                await fhem.CommandDeleteReading(self.hash, self.hash["NAME"] + " line_" + del_reading)

            self._last_data = flat_data
            
        except Exception:
            self.logger.exception("Failed...")
            pass

class WienerlinienAPI:
    """Call API."""

    def __init__(self, stopid):
        """Initialize."""
        self.session = aiohttp.ClientSession()
        self.stopid = stopid

    async def get_json(self):
        """Get json from API endpoint."""
        value = None
        url = BASE_URL.format(self.stopid)
        try:
            response = await self.session.get(url)
            value = await response.json()
        except Exception:
            self.logger.exception("Failed...")
            pass

        return value



import asyncio
import json
import time
import functools
from ring_doorbell import Ring, Auth
from oauthlib.oauth2 import MissingTokenError

from .. import fhem
from .. import utils

class ring:

    def __init__(self, logger):
        self.logger = logger
        self.loop = asyncio.get_event_loop()
        self._username = None
        self._password = ""
        self._token = ""
        self._2facode = None
        self._attr_dingPollInterval = 5
        self._attr_deviceUpdateInterval = 300
        self._history = []
        self._rdevice = None
        self._lastrecording_url = ""
        self._livestreamjson = ""
        self._snapshot = None
        self._attr_list = {
            "deviceUpdateInterval": { "default": 300, "format": "int" },
            "dingPollInterval": { "default": 2, "format": "int" }
        }
        return

    async def token_updated(self, token):
        self._token = token
        encrypted_token = utils.encrypt_string(json.dumps(token), self._reading_encryption_key)
        await fhem.readingsSingleUpdate(self.hash, "token", encrypted_token, 1)        

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        self.hash = hash
        if len(args) < 5:
            return "Usage: define rrring PythonModule ring <USERNAME> <RING_DEVICE_NAME>"
        self._username = args[3]
        self._rdevname = args[4]
        self._reading_encryption_key = await fhem.getUniqueId(hash)
        # ring service
        self._ring = None
        # ring doorbell/chime/... device
        self._rdevice = None
        self.hash["USERNAME"] = args[3]
        self.hash["RINGDEVICE"] = args[4]

        await utils.handle_define_attr(self._attr_list, self, hash)

        asyncio.create_task(self.ring_login())
        return ""

    async def ring_login(self):
        if self._token == "":
            token_reading = await fhem.ReadingsVal(self.hash['NAME'], "token", "")
            if token_reading != "":
                token_reading = utils.decrypt_string(token_reading, self._reading_encryption_key)
                self._token = json.loads(token_reading)
        if self._password == "":
            self._password = await fhem.ReadingsVal(self.hash['NAME'], "password", "")
            if self._password != "":
                self._password = utils.decrypt_string(self._password, self._reading_encryption_key)
        try:
            ret = await utils.run_blocking(functools.partial(self.blocking_login))
            if ret:
                await fhem.readingsSingleUpdate(self.hash, "state", ret, 1)
                return
            await fhem.readingsSingleUpdate(self.hash, "state", "connected", 1)
            # start udpate loop, we are already in a task, therefore no need to create
            await self.update_loop()
        except:
            await fhem.readingsSingleUpdate(self.hash, "state", "Login failed", 1)
            self.logger.error("Login failed")

    async def update_loop(self):
        try:
            await utils.run_blocking(functools.partial(self._ring.update_data))
            devices = await utils.run_blocking(functools.partial(self._ring.devices))
            for dev_type in devices:
                for dev in devices[dev_type]:
                    if dev.name == self._rdevname:
                        self._rdevice = dev

            if self._rdevice is None:
                await fhem.readingsSingleUpdate(self.hash, "state", "device not found", 1)
                return

            await self.update_readings()

            asyncio.create_task(self.update_dings_loop())

            while True:
                try:
                    await utils.run_blocking(functools.partial(self.poll_device))
                    await self.update_readings()
                    # handle history
                    if len(self._history) > 0:
                        i = 1
                        for event in self._history:
                            await self.update_history_readings(event, i)
                            i += 1
                except:
                    self.logger.exception("Failed to poll devices")
                await asyncio.sleep(self._attr_deviceUpdateInterval)
        except:
            self.logger.exception("Failed to update devices")

    async def update_dings_loop(self):
        alert_active = 0
        while True:
            try:
                await utils.run_blocking(functools.partial(self.poll_dings))
                # handle alerts
                alerts = self._ring.active_alerts()
                self.logger.debug("Received dings: " + str(alerts))
                if len(alerts) > 0:
                    alert_active = 1
                    for alert in alerts:
                        await self.update_alert_readings(alert)
                elif alert_active == 1:
                    alert_active = 0
                    await fhem.readingsSingleUpdateIfChanged(self.hash, "state", "connected", 1)
                    await utils.run_blocking(functools.partial(self.poll_device))
                    await self.update_readings()
            except:
                self.logger.exception("Failed to poll dings...")
                await utils.run_blocking(functools.partial(self.blocking_login))
            await asyncio.sleep(self._attr_dingPollInterval)

    async def update_alert_readings(self, alert):
        await fhem.readingsBeginUpdate(self.hash)
        await fhem.readingsBulkUpdate(self.hash, "alert_id", alert["id"])
        await fhem.readingsBulkUpdate(self.hash, "alert_kind", alert["kind"])
        await fhem.readingsBulkUpdate(self.hash, "alert_sip_to", alert["sip_to"])
        await fhem.readingsBulkUpdate(self.hash, "alert_sip_token", alert["sip_token"])
        await fhem.readingsBulkUpdate(self.hash, "state", alert["kind"])
        await fhem.readingsEndUpdate(self.hash, 1)

    async def update_history_readings(self, event, idx):
        await fhem.readingsBeginUpdate(self.hash)
        await fhem.readingsBulkUpdateIfChanged(self.hash, "history_" + str(idx) + "_id", event["id"])
        await fhem.readingsBulkUpdateIfChanged(self.hash, "history_" + str(idx) + "_kind", event["kind"])
        await fhem.readingsBulkUpdateIfChanged(self.hash, "history_" + str(idx) + "_answered", event["answered"])
        await fhem.readingsBulkUpdateIfChanged(self.hash, "history_" + str(idx) + "_created_at", event["created_at"])
        await fhem.readingsEndUpdate(self.hash, 1)

    async def update_readings(self):
        await fhem.readingsBeginUpdate(self.hash)
        await fhem.readingsBulkUpdateIfChanged(self.hash, "address", self._rdevice.address)
        await fhem.readingsBulkUpdateIfChanged(self.hash, "family", self._rdevice.family)
        await fhem.readingsBulkUpdateIfChanged(self.hash, "id", self._rdevice.device_id)
        await fhem.readingsBulkUpdateIfChanged(self.hash, "model", self._rdevice.model)
        await fhem.readingsBulkUpdateIfChanged(self.hash, "firmware", self._rdevice.firmware)
        await fhem.readingsBulkUpdateIfChanged(self.hash, "latitude", self._rdevice.latitude)
        await fhem.readingsBulkUpdateIfChanged(self.hash, "longitude", self._rdevice.longitude)
        await fhem.readingsBulkUpdateIfChanged(self.hash, "kind", self._rdevice.kind)
        await fhem.readingsBulkUpdateIfChanged(self.hash, "name", self._rdevice.name)
        await fhem.readingsBulkUpdateIfChanged(self.hash, "timezone", self._rdevice.timezone)
        await fhem.readingsBulkUpdateIfChanged(self.hash, "wifi_name", self._rdevice.wifi_name)
        await fhem.readingsBulkUpdateIfChanged(self.hash, "wifi_signal_strength", self._rdevice.wifi_signal_strength)
        await fhem.readingsBulkUpdateIfChanged(self.hash, "wifi_signal_category", self._rdevice.wifi_signal_category)
        await self.update_if_available("battery_life")
        await self.update_if_available("existing_doorbell_type")
        await self.update_if_available("existing_doorbell_type_enabled")
        await self.update_if_available("existing_doorbell_type_duration")
        await self.update_if_available("subscribed")
        await self.update_if_available("subscribed_motion")
        await self.update_if_available("has_subscription")
        await self.update_if_available("volume")
        await self.update_if_available("connection_status")
        if self._rdevice.family == "doorbots" or self._rdevice.family == "authorized_doorbots":
            await self.update_if_available("last_recording_id")
            await fhem.readingsBulkUpdateIfChanged(self.hash, "last_recording_url", self._lastrecording_url)
            if self._livestreamjson != "":
                await fhem.readingsBulkUpdateIfChanged(self.hash, "livestream_json", json.dumps(self._livestreamjson))
            # if self._snapshot:
            #     snapshot = '<html><img src="data:image/png,' + self._snapshot + '"/></html>'
            #     await fhem.readingsBulkUpdateIfChanged(self.hash, "snapshot", snapshot)
        await fhem.readingsEndUpdate(self.hash, 1)

    async def update_if_available(self, reading):
        if hasattr(self._rdevice, reading):
            await fhem.readingsBulkUpdateIfChanged(self.hash, reading, getattr(self._rdevice, reading))

    def poll_dings(self):
        self._ring.update_dings()

    def poll_device(self):
        self._rdevice.update_health_data()
        self._history = []
        if self._rdevice.family == "doorbots" or self._rdevice.family == "authorized_doorbots":
            # disable it, as it creates a lot of history entries
            #self._livestreamjson = self._rdevice.live_streaming_json
            for event in self._rdevice.history(limit=5):
                self._history.append(event)
            self._lastrecording_url = self._rdevice.recording_url(self._rdevice.last_recording_id)
            if self._lastrecording_url is False:
                self.blocking_login()
            #self._snapshot = self._rdevice.get_snapshot()

    def blocking_login(self):
        def token_updater(token):
            asyncio.run_coroutine_threadsafe(self.token_updated(token), self.loop).result()

        if self._token != "":
            self._auth = Auth("MyProject/1.0", self._token, token_updater)
        else:
            if self._password != "":
                self._auth = Auth("MyProject/1.0", None, token_updater)
                if self._2facode:
                    self._auth.fetch_token(self._username, self._password, self._2facode)
                else:
                    try:
                        self._auth.fetch_token(self._username, self._password)
                    except MissingTokenError:
                        return "please set 2fa_code"
            else:
                return "please set password"
        self._ring = Ring(self._auth)

    # FHEM FUNCTION
    async def Undefine(self, hash):
        return

    # FHEM FUNCTION
    async def Set(self, hash, args, argsh):
        set_list_conf = {
           "password": { "args": ["password"] },
           "2fa_code": { "args": ["2facode"] }
        }
        if self._rdevice is not None and self._rdevice.has_capability("volume"):
            set_list_conf['volume'] = { "args": ["volume"], "params": { "volume": {"format": "int"}}, "options": "slider,0,1,10"}
        return await utils.handle_set(set_list_conf, self, hash, args, argsh)

    async def set_volume(self, hash, params):
        new_vol = params['volume']
        asyncio.create_task(self.set_volume_task(new_vol))
    
    async def set_volume_task(self, new_volume):
        await utils.run_blocking(functools.partial(self.set_volume_blocking, new_volume))
        await self.update_readings()

    def set_volume_blocking(self, new_volume):
        self._rdevice.volume = new_volume

    async def set_password(self, hash, params):
        self._password = params['password']
        encrypted_password = utils.encrypt_string(self._password, self._reading_encryption_key)
        await fhem.readingsSingleUpdateIfChanged(self.hash, "password", encrypted_password, 1)
        asyncio.create_task(self.ring_login())

    async def set_2fa_code(self, hash, params):
        self._2facode = params['2facode']
        asyncio.create_task(self.ring_login())

    async def Attr(self, hash, args, argsh):
        return await utils.handle_attr(self._attr_list, self, hash, args, argsh)


"""
Starts a service to scan in intervals for new devices.

ATTENTION: This one should only be used for DLNA_DMR devices as the used library netdisco is deprecated!!!
"""
from datetime import timedelta
import json
import logging
import concurrent.futures
import functools
import asyncio
import aiohttp
import sys

from async_upnp_client.aiohttp import AiohttpNotifyServer, AiohttpSessionRequester
from async_upnp_client.advertisement import UpnpAdvertisementListener
from async_upnp_client import UpnpFactory
from async_upnp_client.search import async_search as async_ssdp_search
from ipaddress import IPv4Address

from .. import fhem

# ssdp search class can be used by other modules as well
class ssdp:

    __instance = None

    @staticmethod
    def getInstance(logger):
        if ssdp.__instance is None:
            ssdp.__instance = ssdp(logger)
        return ssdp.__instance

    def __init__(self, logger):
        if ssdp.__instance is not None:
            raise Exception("ssdp already defined, use getInstance")
        self.logger = logger
        self.listeners = []
        self.search_task = None
        self.advertisement_task = None
        self.nr_started_searches = 0

        # build upnp/aiohttp requester
        session = aiohttp.ClientSession()
        self.requester = AiohttpSessionRequester(session, True)
        # create upnp device
        self.factory = UpnpFactory(self.requester)

    async def start_search(self):
        self.nr_started_searches += 1
        self.search_task = asyncio.create_task(self.search())

        if self.advertisement_task is None:
            self.advertisement_task = asyncio.create_task(
                self.advertisements())

    async def stop_search(self):
        self.nr_started_searches -= 1
        # stop search only when last client stops it
        if self.nr_started_searches == 0:
            await self.listener.async_stop()
            if self.search_task:
                self.search_task.cancel()
            if self.advertisement_task:
                self.advertisement_task.cancel()

    def register_listener(self, listener, ssdp_filter={"service_type":"ssdp:all"}):
        listenerFilter = {
            "listener": listener,
            "ssdp_filter": ssdp_filter,
            "found_devices": {}
        }
        self.listeners.append(listenerFilter)

    async def updated_device(self, udn):
        return

    async def create_device(self, msg, data):
        upnp_device = None
        try:
            if msg != "byebye":
                upnp_device = await self.factory.async_create_device(data['location'])
        except:
            # upnp_device remains None
            pass
        return upnp_device

    async def handle_msg(self, msg, data):
        try:
            data = {key.lower(): str(value) for key, value in data.items()}
            usn = data['usn']
            arr = usn.split("::")
            udn = arr[0]
            if len(arr) > 1:
                st = arr[1]
            else:
                st = ""
            self.logger.debug("found: " + usn)
            # filter for listeners
            for listenerFilter in self.listeners:
                listener = listenerFilter['listener']
                ssdp_filter = listenerFilter['ssdp_filter']
                filter_udn = None
                filter_st = None
                if "udn" in ssdp_filter:
                    filter_udn = ssdp_filter['udn']
                if "service_type" in ssdp_filter:
                    filter_st = ssdp_filter['service_type']
                if ((udn == filter_udn or filter_udn is None) and
                (st == filter_st or filter_st == "ssdp:all")):
                    if usn not in listenerFilter['found_devices'] and msg == "alive":
                        self.logger.debug("create device: " + usn)
                        listenerFilter['found_devices'][usn] = await self.create_device(msg, data)
                        if listenerFilter['found_devices'][usn]:
                            self.logger.debug("found device: " + usn)
                            await listener.found_device(listenerFilter['found_devices'][usn])
                    elif usn in listenerFilter['found_devices'] and msg == "byebye":
                        self.logger.debug("removed device: " + usn)
                        await listener.removed_device(listenerFilter['found_devices'][usn])
                        del listenerFilter['found_devices'][usn]
        except:
            self.logger.exception("Error in handle_msg")

    async def search(self):
        await self.search_once()


    async def search_once(self):
        timeout = 30
        service_type = "ssdp:all"
        source_ip = None
        if sys.platform == 'win32' and not source_ip:
            self.logger.debug(
                'Running on win32 without --bind argument, forcing to "0.0.0.0"')
            source_ip = '0.0.0.0'  # force to IPv4 to prevent asyncio crash/WinError 10022
        if source_ip:
            source_ip = IPv4Address(source_ip)

        async def on_response(data):
            await self.handle_msg("alive", data)

        try:
            await async_ssdp_search(service_type=service_type,
                                    source_ip=source_ip,
                                    timeout=timeout,
                                    async_callback=on_response)
        except asyncio.CancelledError:
            self.logger.debug("search cancelled")

    async def advertisements(self):
        """Listen for advertisements."""
        try:
            source_ip = None
            if sys.platform == 'win32' and not source_ip:
                self.logger.debug(
                    'Running on win32 without --bind argument, forcing to "0.0.0.0"')
                source_ip = '0.0.0.0'  # force to IPv4 to prevent asyncio crash/WinError 10022
            if source_ip:
                source_ip = IPv4Address(source_ip)

            async def on_alive(data):
                await self.handle_msg("alive", data)

            async def on_byebye(data):
                await self.handle_msg("byebye", data)

            async def on_update(data):
                return

            self.listener = UpnpAdvertisementListener(on_alive=on_alive,
                                                    on_byebye=on_byebye,
                                                    on_update=on_update,
                                                    source_ip=source_ip)
            await self.listener.async_start()
        except:
            self.logger.exception("Error in advertisements listener")


class discover_upnp:

    def __init__(self, logger):
        self.logger = logger
        self.hash = None
        self.loop = asyncio.get_event_loop()
        self.create_devs = {}

    async def found_device(self, upnp_device):
        if upnp_device and upnp_device.udn:
            await fhem.readingsSingleUpdate(self.hash, upnp_device.udn.replace(":",".") + "_" + upnp_device.device_type.replace(":","."), upnp_device.friendly_name, 0)
            if upnp_device.device_type == "urn:schemas-upnp-org:device:MediaRenderer:1":
                self.create_devs[upnp_device.udn] = {
                    "name": "".join(filter(str.isalnum, upnp_device.friendly_name)) + "_" + upnp_device.device_type.split(":")[-2],
                    "devname": "".join(filter(str.isalnum, upnp_device.friendly_name)) + "_" + upnp_device.device_type.split(":")[-2]
                }
        # if upnp_device.device_type == "urn:schemas-upnp-org:device:MediaRenderer:1":
        #     if not (await fhem.checkIfDeviceExists(self.hash, "PYTHONTYPE", "dlna_dmr", "UDN", upnp_device.udn)):
        #         devname = devname = upnp_device.friendly_name + "_" + upnp_device.model_name
        #         devname = ''.join(filter(str.isalnum, devname))
        #         await fhem.CommandDefine(self.hash, devname + " PythonModule dlna_dmr " + upnp_device.udn)

    async def removed_device(self, upnp_device):
        return

    # FHEM Define
    async def Define(self, hash, args, argsh):
        """Start a discovery service."""
        self.hash = hash
        ssdp.getInstance(self.logger).register_listener(self)
        await ssdp.getInstance(self.logger).start_search()
        await fhem.readingsSingleUpdate(hash, "state", "active", 0)

        if await fhem.AttrVal(self.hash['NAME'], "icon", "") == "":
            await fhem.CommandAttr(self.hash, self.hash["NAME"] + " icon rc_SEARCH")

        return ""
    
    # FHEM Set
    async def Set(self, hash, args, argsh):
        if (len(args) < 2 or args[1] == "?"):
            set_devs = []
            for dev in self.create_devs:
                set_devs.append(self.create_devs[dev]["name"])
            set_list = ""
            if len(set_devs) > 0:
                set_list = "create:" + ",".join(set_devs)
            return ("Unknown argument ?, choose one of " + set_list)
        else:
            cmd = args[1]
            if cmd == "create":
                for udn in self.create_devs:
                    if self.create_devs[udn]["name"] == args[2]:
                        await fhem.CommandDefine(self.hash, self.create_devs[udn]["devname"] + " PythonModule dlna_dmr " + udn)

    # FHEM Undefine
    async def Undefine(self, hash):
        ssdp.getInstance(self.logger).stop_search()



import asyncio
import websockets
import json
import traceback
import logging
import concurrent.futures
import functools
import site
import sys
import importlib
import time
from . import fhem
from . import pkg_installer


logger = logging.getLogger(__name__)

loadedModuleInstances = {}
moduleLoadingRunning = {}
wsconnection = None

pip_lock = asyncio.Lock()

connection_start = 0
fct_timeout = 60

def getFhemPyDeviceByName(name):
    if name in loadedModuleInstances:
        return loadedModuleInstances[name]
    return None

async def pybinding(websocket, path):
    global connection_start
    connection_start = time.time()
    logger.info("FHEM connection started: " + websocket.remote_address[0])
    pb = PyBinding(websocket)
    fhem.updateConnection(pb)
    await fhem.send_version()
    try:
        async for message in websocket:
            asyncio.create_task(pb.onMessage(message))
    except websockets.exceptions.ConnectionClosedError:
        logger.error("Connection closed error", exc_info=True)
        logger.info("Restart binding")
        sys.exit(1)

class PyBinding:

    msg_listeners = []

    def __init__(self, websocket):
        self.wsconnection = websocket

    def registerMsgListener(self, listener, awaitid):
        self.msg_listeners.append({"func": listener, "awaitId": awaitid})

    async def send(self, msg):
        await self.wsconnection.send(msg.encode("utf-8"))

    async def sendBackReturn(self, hash, ret):
        retHash = hash.copy()
        retHash['finished'] = 1
        retHash['returnval'] = ret
        retHash['id'] = hash['id']
        msg = json.dumps(retHash)
        logger.debug("<<< WS: " + msg)
        await self.wsconnection.send(msg.encode("utf-8"))
        fhem.setFunctionInactive(hash)        

    async def sendBackError(self, hash, error):
        logger.error(error + "(id: {})".format(hash['id']))
        retHash = hash.copy()
        retHash['finished'] = 1
        retHash['error'] = error
        retHash['id'] = hash['id']
        msg = json.dumps(retHash, ensure_ascii=False)
        logger.debug("<<< WS: " + msg)
        await self.wsconnection.send(msg.encode("utf-8"))
        fhem.setFunctionInactive(hash)

    async def updateHash(self, hash):
        retHash = hash.copy()
        retHash['msgtype'] = "update_hash"
        del retHash['id']
        msg = json.dumps(retHash, ensure_ascii=False)
        logger.debug("<<< WS: " + msg)
        await self.wsconnection.send(msg.encode("utf-8"))

    def getLogLevel(self, verbose_level):
        if verbose_level == "5":
            return logging.DEBUG
        elif verbose_level == "4":
            return logging.INFO
        elif verbose_level == "3":
            return logging.WARNING
        else:
            return logging.ERROR

    async def onMessage(self, payload):
        try:
            await self._onMessage(payload)
        except SystemExit as se:
            sys.exit(1)
        except:
            logger.exception("Failed to handle message: " + str(payload))

    async def _onMessage(self, payload):
        msg = payload
        logger.debug(">>> WS: " + msg)
        hash = None
        try:
            hash = json.loads(msg)
        except:
            logger.error("Websocket JSON couldn't be decoded")
            return

        global fct_timeout, connection_start
        if time.time() - connection_start > 120:
            fct_timeout = 10

        try:
            if ("awaitId" in hash and len(self.msg_listeners) > 0):
                removeElement = None
                for listener in self.msg_listeners:
                    if (listener['awaitId'] == hash["awaitId"]):
                        listener['func'](msg)
                        removeElement = listener
                if (removeElement):
                    self.msg_listeners.remove(removeElement)
            else:
                ret = ''
                if hash['msgtype'] == "update":
                    await pkg_installer.force_update_package("fhempy")
                    sys.exit(1)
                if (hash['msgtype'] == "function"):
                    # this is needed to avoid 2 replies on dep installation
                    fhem_reply_done = False
                    fhem.setFunctionActive(hash)
                    # load module
                    nmInstance = None
                    if hash['function'] == "Rename":
                        if hash['NAME'] in loadedModuleInstances:
                            loadedModuleInstances[hash['args'][1]] = loadedModuleInstances[hash['args'][0]]
                            del loadedModuleInstances[hash['args'][0]]
                            self.sendBackReturn(hash, "")
                            return 0

                    if (hash['function'] != "Undefine"):
                        # Load module and execute Define if Define isn't called right now
                        if (not (hash["NAME"] in loadedModuleInstances)):
                            if hash["NAME"] in moduleLoadingRunning:
                                await self.sendBackReturn(hash, "")
                                return 0

                            moduleLoadingRunning[hash["NAME"]] = True

                            # loading a module might take some time, therefore sendBackReturn now
                            await self.sendBackReturn(hash, "")
                            fhem_reply_done = True

                            try:
                                # check dependencies
                                deps_ok = pkg_installer.check_dependencies(hash["PYTHONTYPE"])
                                if deps_ok == False:
                                    # readingsSingleUpdate inform about dep installation
                                    await fhem.readingsSingleUpdate(hash, "state", "Installing updates...", 1)
                                    # run only one installation and do depcheck before any other installation
                                    async with pip_lock:
                                        # make sure that all import caches are up2date before check
                                        importlib.invalidate_caches()
                                        # check again if something changed for dependencies
                                        deps_ok = pkg_installer.check_dependencies(hash["PYTHONTYPE"])
                                        if deps_ok == False:
                                            # start installation in a separate asyncio thread
                                            await pkg_installer.check_and_install_dependencies(hash["PYTHONTYPE"])
                                            # update cache again after install
                                            if not site.getusersitepackages() in sys.path:
                                                logger.debug("add pip path: " + site.getusersitepackages())
                                                sys.path.append(site.getusersitepackages())
                                            importlib.invalidate_caches()
                                    # when installation finished, inform user
                                    await fhem.readingsSingleUpdate(hash, "state", "Installation finished. Define now...", 1)
                                    # wait 5s so that user can read the msg about installation
                                    await asyncio.sleep(5)
                                    # continue define

                                # import module
                                pymodule = "lib." + hash["PYTHONTYPE"] + "." + hash["PYTHONTYPE"]
                                module_object = importlib.import_module(pymodule)
                                # create instance of class with logger
                                target_class = getattr(module_object, hash["PYTHONTYPE"])
                                moduleLogger = logging.getLogger(hash["NAME"])
                                moduleLogger.setLevel(self.getLogLevel(await fhem.AttrVal(hash["NAME"], "verbose", "3")))
                                loadedModuleInstances[hash["NAME"]] = target_class(moduleLogger)
                                del moduleLoadingRunning[hash["NAME"]]
                                if (hash["function"] != "Define"):
                                    func = getattr(loadedModuleInstances[hash["NAME"]], "Define", "nofunction")
                                    await asyncio.wait_for(func(hash, hash['defargs'], hash['defargsh']), fct_timeout)
                            except asyncio.TimeoutError:
                                errorMsg = f"Function execution >{fct_timeout}s, cancelled: {hash['NAME']} Define"
                                if fhem_reply_done:
                                    await fhem.readingsSingleUpdate(hash, "state", errorMsg, 1)
                                else:
                                    await self.sendBackError(hash, errorMsg)
                                return 0
                            except ModuleNotFoundError:
                                errorMsg = f"Module failed to load: {hash['PYTHONTYPE']}\nMaybe you need to update fhempy on this or remote peer."
                                errorMsg += "\n\nStacktrace:\n" + traceback.format_exc()
                                if fhem_reply_done:
                                    await fhem.readingsSingleUpdate(hash, "state", errorMsg, 1)
                                else:
                                    await self.sendBackError(hash, errorMsg)
                                return 0
                            except Exception:
                                errorMsg = "Failed to load module " + hash["PYTHONTYPE"] + ": " + traceback.format_exc()
                                if fhem_reply_done:
                                    await fhem.readingsSingleUpdate(hash, "state", errorMsg, 1)
                                else:
                                    await self.sendBackError(hash, errorMsg)
                                return 0
                        
                    nmInstance = loadedModuleInstances[hash["NAME"]]

                    if (nmInstance != None):
                        try:
                            # handle verbose level of logging
                            if hash["function"] == "Attr" and hash["args"][2] == "verbose":
                                moduleLogger = logging.getLogger(hash["NAME"])
                                if hash["args"][0] == "set":
                                    moduleLogger.setLevel(self.getLogLevel(hash["args"][3]))
                                else:
                                    moduleLogger.setLevel(logging.ERROR)
                            else:
                                # call Set/Attr/Define/...
                                func = getattr(nmInstance, hash["function"], "nofunction")
                                if (func != "nofunction"):
                                    logger.debug(f"Start function {hash['NAME']}:{hash['function']}")
                                    if hash["function"] == "Undefine":
                                        try:
                                            ret = await asyncio.wait_for(func(hash), fct_timeout)
                                        except:
                                            del loadedModuleInstances[hash["NAME"]]
                                    else:
                                        ret = await asyncio.wait_for(func(hash, hash['args'], hash['argsh']), fct_timeout)
                                    logger.debug(f"End function {hash['NAME']}:{hash['function']}")
                                    if (ret == None):
                                        ret = ""
                                    if fhem_reply_done:
                                        await self.updateHash(hash)
                        except asyncio.TimeoutError:
                            errorMsg = f"Function execution >{fct_timeout}s, cancelled: {hash['NAME']} - {hash['function']}"
                            if fhem_reply_done:
                                await fhem.readingsSingleUpdate(hash, "state", errorMsg, 1)
                            else:
                                await self.sendBackError(hash, errorMsg)
                            return 0
                        except:
                            errorMsg = "Failed to execute function " + hash["function"] + ": " + traceback.format_exc()
                            if fhem_reply_done:
                                await fhem.readingsSingleUpdate(hash, "state", errorMsg, 1)
                            else:
                                await self.sendBackError(hash, errorMsg)
                            return 0
                    
                    if (hash['function'] == "Undefine"):
                        if hash["NAME"] in loadedModuleInstances:
                            del loadedModuleInstances[hash["NAME"]]
                    
                    if fhem_reply_done is False:
                        await self.sendBackReturn(hash, ret)

        except SystemExit as se:
            raise se
        except Exception:
            logger.error("Failed to handle message: ", exc_info=True)

def run():
    logger.info("Starting pythonbinding...")
    asyncio.get_event_loop().run_until_complete(
        websockets.serve(pybinding, '0.0.0.0', 15733, ping_timeout=None, ping_interval=None))
    asyncio.get_event_loop().run_forever()




import asyncio
import logging
from bleak import discover

from .. import fhem

class discover_ble:

    def __init__(self, logger):
        self.logger = logger
        # disable bleak discovery messages
        logging.getLogger("bleak.backends.bluezdbus.discovery").setLevel(logging.ERROR)
        self.hash = None
        self.blescanTask = None
        return

    async def runBleScan(self):
        while True:
            try:
                devices = await discover()
                for d in devices:
                    if d.name == "GfBT Project":
                        if not await fhem.checkIfDeviceExists(self.hash, "TYPE", "GFPROBT", "MAC", d.address):
                            self.logger.debug("create device: " + d.name + " / " + d.address + " / rssi: " + str(d.rssi))
                            await fhem.CommandDefine(self.hash, d.name + "_" + d.address.replace(":", "") +  " GFPROBT '" + d.address + "'")
                        else:
                            self.logger.debug("existing device: " + d.name + " / " + d.address + " / rssi: " + str(d.rssi))
                    elif d.name == "CC-RT-BLE":
                        if not await fhem.checkIfDeviceExists(self.hash, "PYTHONTYPE", "eq3bt", "MAC", d.address):
                            self.logger.debug("create device: " + d.name + " / " + d.address + " / rssi: " + str(d.rssi))
                            await fhem.CommandDefine(self.hash, d.name + "_" + d.address.replace(":", "") +  " PythonModule eq3bt '" + d.address + "'")
                        else:
                            self.logger.debug("existing device: " + d.name + " / " + d.address + " / rssi: " + str(d.rssi))
                    elif d.name[0:7] == "Expert_":
                        if not await fhem.checkIfDeviceExists(self.hash, "PYTHONTYPE", "nespresso_ble", "MAC", d.address):
                            self.logger.debug("create device: " + d.name + " / " + d.address + " / rssi: " + str(d.rssi))
                            await fhem.CommandDefine(self.hash, d.name + "_" + d.address.replace(":", "") +  " PythonModule nespresso_ble '" + d.address + "'")
                        else:
                            self.logger.debug("existing device: " + d.name + " / " + d.address + " / rssi: " + str(d.rssi))
                    else:
                        self.logger.debug("found unhandled device: " + d.name + ", " + d.address + ", rssi: " + str(d.rssi))
            except:
                self.logger.error("BLE Scan failed, retry in 600s", exc_info=True)
            await asyncio.sleep(600)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        self.hash = hash

        await fhem.readingsBeginUpdate(hash)
        await fhem.readingsBulkUpdateIfChanged(hash, "state", "active")
        await fhem.readingsEndUpdate(hash, 1)

        if await fhem.AttrVal(self.hash['NAME'], "icon", "") == "":
            await fhem.CommandAttr(self.hash, self.hash["NAME"] + " icon rc_SEARCH")

        if self.blescanTask:
            self.blescanTask.cancel()

        self.blescanTask = asyncio.create_task(self.runBleScan())

        return ""

    # FHEM FUNCTION
    async def Undefine(self, hash):
        if self.blescanTask:
            self.blescanTask.cancel()
        return



import asyncio

from .. import fhem
from .. import utils
from ..generic import FhemModule

class helloworld(FhemModule):

    def __init__(self, logger):
        super().__init__(logger)
        
        attr_config = {
            "interval": {"default": 100, "format": "int", "help": "Change interval, default is 100."}
        }
        self.set_attr_config(attr_config)

        set_config = {
            "mode": { 
                "args": ["mode"],
                "argsh": ["mode"],
                "params": {
                    "mode": { "default": "eco", "optional": False }
                },
                "options": "eco,comfort"
            },
            "desiredTemp": {
                "args": ["temperature"],
                "options": "slider,10,1,30"
            },
            "holidayMode": { 
                "args": ["start", "end", "temperature"],
                "params": {
                    "start": {"default": "Monday"},
                    "end": {"default": "23:59"},
                    "temperature": {"default": ""}
                }
            },
            "on": { 
                "args": ["seconds"], 
                "params": { 
                    "seconds": { "default": "", "optional": True}
                },
                "help": "Specify seconds as parameter to change to off after X seconds."
            },
            "off": {}
        }
        self.set_set_config(set_config)
        return

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        await fhem.readingsBeginUpdate(hash)
        await fhem.readingsBulkUpdateIfChanged(hash, "state", "on")
        await fhem.readingsEndUpdate(hash, 1)
        return ""

    # Attribute function format: set_attr_NAMEOFATTRIBUTE(self, hash)
    # self._attr_NAMEOFATTRIBUTE contains the new state
    async def set_attr_interval(self, hash):
        # attribute was set to self._attr_interval
        # you can use self._attr_interval already with the new variable
        pass

    # Set functions in format: set_NAMEOFSETFUNCTION(self, hash, params)
    async def set_on(self, hash, params):
        # params contains the keyword which was defined in set_list_conf for "on"
        # if not provided by the user it will be "" as defined in set_list_conf (default = "" and optional = True)
        seconds = params['seconds']
        await fhem.readingsSingleUpdate(hash, "state", "on " + seconds, 1)
        return ""

    async def set_off(self, hash):
        # no params argument here, as set_off doesn't have arguments defined in set_list_conf
        await fhem.readingsSingleUpdate(hash, "state", "off", 1)
        self.create_async_task(self.long_running_task())
        return ""
    
    async def long_running_task(self):
        await asyncio.sleep(30)
        await fhem.readingsSingleUpdate(self.hash, "state", "long running off", 1)

    async def set_mode(self, hash, params):
        # user can specify mode as mode=eco or just eco as argument
        # params['mode'] contains the mode provided by user
        mode = params['mode']
        await fhem.readingsSingleUpdate(hash, "mode", mode, 1)
        return ""

    async def set_desiredTemp(self, hash, params):
        temp = params['temperature']
        await fhem.readingsSingleUpdate(hash, "mode", temp, 1)
        return ""

    async def set_holidayMode(self, hash, params):
        start = params['start']
        end = params['end']
        temp = params['temperature']
        await fhem.readingsSingleUpdate(hash, "start", start, 1)
        await fhem.readingsSingleUpdate(hash, "end", end, 1)
        await fhem.readingsSingleUpdate(hash, "temp", temp, 1)
        return ""




import json
import random
import asyncio
import logging
import traceback
import concurrent.futures
import websockets

from .version import __version__

logger = logging.getLogger(__name__)
logger.setLevel(logging.ERROR)

function_active = []
update_locks = {}
wsconnection = None

# TODO use run_coroutine_threadsafe if asyncio.get_event_loop() == None
# this would make all functions threadsafe

def updateConnection(ws):
    global wsconnection
    wsconnection = ws

def setFunctionActive(hash):
    function_active.append(hash["NAME"])

def setFunctionInactive(hash):
    element = function_active.pop()
    if element != hash["NAME"]:
        logger.error(f"Set wrong function inactive, tried {hash['NAME']}, current function_active: {function_active},{element}")

async def getUniqueId(hash):
    cmd = "getUniqueId()"
    return await sendCommandHash(hash, cmd)

async def ReadingsVal(name, reading, default):
    cmd = "ReadingsVal('" + name + "', '" + reading + "', '" + default + "')"
    return await sendCommandName(name, cmd)

async def AttrVal(name, attr, default):
    cmd = "AttrVal('" + name + "', '" + attr + "', '" + default + "')"
    return await sendCommandName(name, cmd)

async def InternalVal(name, internal, default):
    cmd = "InternalVal('" + name + "', '" + internal + "', '" + default + "')"
    return await sendCommandName(name, cmd)

async def addToDevAttrList(name, attr_list):
    cmd = "addToDevAttrList('" + name + "', '" + attr_list + "')"
    return await sendCommandName(name, cmd)

async def setDevAttrList(name, attr_list):
    attr_list += " IODev"
    cmd = "setDevAttrList('" + name + "', '" + attr_list + " '.$readingFnAttributes)"
    return await sendCommandName(name, cmd)

async def readingsBeginUpdate(hash):
    if hash["NAME"] not in update_locks:
        update_locks[hash["NAME"]] = asyncio.Lock()
    await update_locks[hash["NAME"]].acquire()
    cmd = "readingsBeginUpdate($defs{'" + hash["NAME"] + "'});;"
    return await sendCommandHash(hash, cmd)

async def readingsBulkUpdateIfChanged(hash, reading, value):
    value = convertValue(value)
    cmd = "readingsBulkUpdateIfChanged($defs{'" + hash["NAME"] + "'},'" + \
        reading + "','" + value.replace("'", "\\'") + "');;"
    return await sendCommandHash(hash, cmd)

async def readingsBulkUpdate(hash, reading, value, changed=None):
    value = convertValue(value)
    if changed is None:
        cmd = "readingsBulkUpdate($defs{'" + hash["NAME"] + "'},'" + \
            reading + "','" + value.replace("'", "\\'") + "');;"
    else:
        cmd = "readingsBulkUpdate($defs{'" + hash["NAME"] + "'},'" + \
            reading + "','" + value.replace("'", "\\'") + "', " + str(changed) + ");;"
    return await sendCommandHash(hash, cmd)

async def readingsEndUpdate(hash, do_trigger):
    cmd = "readingsEndUpdate($defs{'" + hash["NAME"] + "'}," + str(do_trigger) + ");;"
    res = await sendCommandHash(hash,cmd)
    update_locks[hash["NAME"]].release()
    return res

async def readingsSingleUpdate(hash, reading, value, do_trigger):
    if hash["NAME"] not in update_locks:
        update_locks[hash["NAME"]] = asyncio.Lock()
    async with update_locks[hash["NAME"]]:
        value = convertValue(value)
        cmd = "readingsSingleUpdate($defs{'" + hash["NAME"] + "'},'" + \
            reading + "','" + value.replace("'", "\\'") + "'," + str(do_trigger) + ")"
        return await sendCommandHash(hash, cmd)

async def readingsSingleUpdateIfChanged(hash, reading, value, do_trigger):
    if hash["NAME"] not in update_locks:
        update_locks[hash["NAME"]] = asyncio.Lock()
    async with update_locks[hash["NAME"]]:
        value = convertValue(value)
        cmd = "readingsBeginUpdate($defs{'" + hash["NAME"] + "'});;readingsBulkUpdateIfChanged($defs{'" + hash["NAME"] + "'},'" + \
            reading + "','" + value.replace("'", "\\'") + \
            "');;readingsEndUpdate($defs{'" + \
            hash["NAME"] + "'}," + str(do_trigger) + ");;"
        return await sendCommandHash(hash, cmd)

async def CommandDefine(hash, definition):
    cmd = "CommandDefine(undef, \"" + definition + "\")"
    return await sendCommandHash(hash, cmd)

async def CommandAttr(hash, attrdef):
    cmd = "CommandAttr(undef, \"" + attrdef + "\")"
    return await sendCommandHash(hash, cmd)

async def CommandDeleteReading(hash, deldef):
    cmd = "CommandDeleteReading(undef, \"" + deldef + "\")"
    return await sendCommandHash(hash, cmd)

async def checkIfDeviceExists(hash, typeinternal, typevalue, internal, value):
    cmd = "foreach my $fhem_dev (sort keys %main::defs) {" + \
        "  return 1 if(defined($main::defs{$fhem_dev}{" + typeinternal + "}) && $main::defs{$fhem_dev}{" + typeinternal + "} eq '" + typevalue + "' && $main::defs{$fhem_dev}{" + internal + "} eq '" + value + "');;" + \
        "}" + \
        "return 0;;"
    return await sendCommandHash(hash, cmd)

# UTILS FUNCTIONS TO SEND COMMAND TO FHEM
def convertValue(value):
    if (value == None):
        value = ""
    if (value == True):
        value = 1
    if (value == False):
        value = 0

    return str(value)

async def send_version():
    msg = {
        "msgtype": "version",
        "version": __version__,
    }
    msg = json.dumps(msg, ensure_ascii=False)
    logger.debug("<<< WS: " + msg)
    global wsconnection
    await wsconnection.send(msg)

async def send_and_wait(name, cmd):
    fut = asyncio.get_running_loop().create_future()
    msg = {
        "awaitId": random.randint(10000000, 99999999),
        "NAME": name,
        "msgtype": "command",
        "command": cmd
    }

    def listener(rmsg):
        try:
            fut.set_result(rmsg)
        except:
            logger.error("Failed to set result, received: " + rmsg)

    global wsconnection
    wsconnection.registerMsgListener(listener, msg['awaitId'])
    msg = json.dumps(msg, ensure_ascii=False)
    logger.debug("<<< WS: " + msg)
    try:
        await wsconnection.send(msg)
        logger.debug("message sent successfully")
    except websockets.exceptions.ConnectionClosed:
        logger.error("Connection closed, can't send message.")
    except Exception as e:
        logger.error("Failed to send message via websocket: " + e)
        fut.set_exception(Exception("Failed to send message via websocket"))
    
    return await fut


async def sendCommandName(name, cmd, hash=None):
    ret = ""
    try:
        logger.debug("sendCommandName START")
        while len(function_active) != 0:
            if function_active[-1] == name:
                break
            await asyncio.sleep(0.001)
        # wait max 1s for reply from FHEM
        jsonmsg = await asyncio.wait_for(send_and_wait(name, cmd), 15)
        logger.debug("sendCommandName END")
        ret = json.loads(jsonmsg)['result']
    except asyncio.TimeoutError:
        logger.error("Timeout - NO RESPONSE for command: " + cmd)
        ret = ""
    except concurrent.futures.CancelledError:
        # function timeout
        pass
    except Exception as e:
        logger.error("Exception while waiting for reply: " + e)
        traceback.format_exc()
        ret = str(e)
    
    return ret

async def sendCommandHash(hash, cmd):
    return await sendCommandName(hash["NAME"], cmd, hash)




import asyncio
import logging
import concurrent.futures
from cryptography.fernet import Fernet
from codecs import encode, decode
from functools import reduce
import base64
from . import fhem

def encrypt_string(plain_text, fhem_unique_id):
  key = base64.b64encode(fhem_unique_id.encode('utf-8'))
  cipher_suite = Fernet(key)
  encrypted_text = cipher_suite.encrypt(plain_text.encode("utf-8"))
  return reduce(encode, ('zlib', 'base64'),encrypted_text).decode("utf-8")

def decrypt_string(encrypted_text, fhem_unique_id):
  key = base64.b64encode(fhem_unique_id.encode('utf-8'))
  encrypted_text = encrypted_text.encode("utf-8")
  uncompressed_text = reduce(decode, ('base64', 'zlib'),encrypted_text)
  cipher_suite = Fernet(key)
  return cipher_suite.decrypt(uncompressed_text).decode("utf-8")

async def run_blocking(function):
  try:
    with concurrent.futures.ThreadPoolExecutor() as pool:
      return await asyncio.get_event_loop().run_in_executor(
          pool, function)
  except:
    logging.getLogger(__name__).exception("Error in asyncio thread")
    raise

def run_blocking_task(function):
  return asyncio.create_task(run_blocking(function))

# example config
# attr_list = {
#   "attribute1": {"default": 10, "format": "int"}
# }
async def handle_attr(attr_list, obj, hash, args, argsh):
  cmd = args[0]
  name = args[1]
  attr_name = args[2]
  attr_val = args[3]
  if attr_name in attr_list:
    if cmd == "set":
      setattr(obj, "_attr_" + attr_name, convert2format(attr_val, attr_list[attr_name]))
    else:
      setattr(obj, "_attr_" + attr_name, attr_list[attr_name]['default'])

  # call set_attr_....
  fct_name = "set_attr_" + attr_name
  try:
    fct_call = getattr(obj, fct_name)
    return await fct_call(hash)
  except AttributeError:
    pass

  return

async def handle_define_attr(attr_list, obj, hash):
  add_to_list = []
  for attr in attr_list:
    if 'options' in attr_list[attr]:
      attr_opt = attr + ":" + attr_list[attr]['options']
    else:
      attr_opt = attr
    add_to_list.append(attr_opt)
  await fhem.setDevAttrList(hash["NAME"], " ".join(add_to_list))

  for attr in attr_list:
    curr_val = await fhem.AttrVal(hash['NAME'], attr, "")
    if curr_val == "":
      curr_val = attr_list[attr]['default']
    setattr(obj, "_attr_" + attr, convert2format(curr_val, attr_list[attr]))

    # call set_attr_....
    fct_name = "set_attr_" + attr
    try:
      fct_call = getattr(obj, fct_name)
      return await fct_call(hash)
    except AttributeError:
      pass

  return

def flatten_json(y):
    out = {}

    def flatten(x, name=''):
        if type(x) is dict:
            for a in x:
                flatten(x[a], name + a + '_')
        elif type(x) is list:
            i = 0
            for a in x:
                flatten(a, name + str(i) + '_')
                i += 1
        else:
            out[name[:-1]] = x

    flatten(y)
    return out

def convert2format(value, list_def):
  if "format" in list_def:
    target_format = list_def['format']
  else:
    target_format = "str"

  if target_format == "int":
    return int(value)
  elif target_format == "float":
    return float(value)
  elif target_format == "str":
    return str(value)
  return value

# example config
# set_list_conf = {
#    "mode": { "args": ["mode"], "argsh": ["mode"], "params": { "mode": { "default": "eco", "optional": False }}, "options": "eco,comfort" },
#    "desiredTemp": { "args": ["temperature"], "options": "slider,10,1,30"},
#    "holidayMode": { "args": ["start", "end", "temperature"], "params": { "start": {"default": "Monday"}, "end": {"default": "23:59"}}},
#    "on": { "args": ["seconds"], "params": { "seconds": {"optional": True}}},
#    "off": {}
# }
async def handle_set(set_list_conf, obj, hash, args, argsh):
  fhem_set_list = []
  if len(args) < 2 or (len(argsh) == 0 and args[1] == "?"): 
    for cmd in set_list_conf:
      if "options" in set_list_conf[cmd]:
        fhem_options = ":" + set_list_conf[cmd]["options"]
      elif "args" in set_list_conf[cmd] or "argsh" in set_list_conf[cmd]:
        fhem_options = ""
      else:
        fhem_options = ":noArg"
      fhem_set_list.append(cmd + fhem_options)
    return "Unknown argument ?, choose one of " + " ".join(fhem_set_list)
  else:
    # get cmd
    cmd = args[1]
    if cmd in set_list_conf:
      all_args = {}
      if "argsh" in set_list_conf[cmd]:
        all_args = set_list_conf[cmd]["argsh"]
      cmd_def = set_list_conf[cmd]
      # map arguments to params
      # add args to all_args
      if ("args" in cmd_def and (len(args) - 2) > len(cmd_def["args"])) or (len(args) > 2 and args[2] == "?"):
        return f"Usage: set {hash['NAME']} {cmd} " + " ".join(cmd_def["args"])
      i=0
      for arg in args[2:]:
        # arg ... mode
        # all_args[mode] = mode argument
        all_args[cmd_def["args"][i]] = arg
        i+=1
      # get default values for other params
      final_params = all_args
      if "params" in cmd_def:
        # add value to params
        for arg in all_args:
          if arg in cmd_def["params"]:
            cmd_def["params"][arg]["value"] = all_args[arg]
        for param in cmd_def["params"]:
          # check if value is available or default value
          # check if all required params are availble
          if "default" in cmd_def["params"][param] and "value" not in cmd_def["params"][param]:
            final_params[param] = cmd_def["params"][param]["default"]
          elif "value" in cmd_def["params"][param]:
            final_params[param] = cmd_def["params"][param]["value"]
          elif "optional" not in cmd_def["params"][param] or cmd_def["params"][param]["optional"] is False:
            # no value found, check if optional
            return f"Required argument {param} missing."
          if "format" in cmd_def["params"][param]:
            final_params[param] = convert2format(final_params[param], cmd_def["params"][param])

      # call function with params
      if "function" in set_list_conf[cmd]:
        fct_name = set_list_conf[cmd]['function']
        final_params['cmd'] = cmd
      else:
        fct_name = "set_" + cmd
      fct_call = getattr(obj, fct_name)
      if len(final_params) > 0:
        return await fct_call(hash, final_params)
      
      return await fct_call(hash)
    else:
      return f"Command not available for this device."


import binascii
import ctypes
import logging
import struct
import time
from collections import namedtuple
from datetime import datetime
from uuid import UUID

import pygatt
from pygatt.exceptions import BLEError, NotConnectedError, NotificationTimeout

_LOGGER = logging.getLogger("nespresso_ble")

# Use full UUID since we do not use UUID from bluepy.btle
CHAR_UUID_DEVICE_NAME = UUID('00002a00-0000-1000-8000-00805f9b34fb')
CHAR_UUID_MANUFACTURER_NAME = UUID('00002a00-0000-1000-8000-00805f9b34fb')
CHAR_UUID_STATE = UUID('06aa3a12-f22a-11e3-9daa-0002a5d5c51b')
CHAR_UUID_NBCAPS = UUID('06aa3a15-f22a-11e3-9daa-0002a5d5c51b')
CHAR_UUID_SLIDER = UUID('06aa3a22-f22a-11e3-9daa-0002a5d5c51b')
CHAR_UUID_WATER_HARDNESS = UUID('06aa3a44-f22a-11e3-9daa-0002a5d5c51b')
CHAR_UUID_AUTH = UUID('06aa3a41-f22a-11e3-9daa-0002a5d5c51b')

Characteristic = namedtuple('Characteristic', ['uuid', 'name', 'format'])

c_uint8 = ctypes.c_uint8

manufacturer_characteristics = Characteristic(CHAR_UUID_MANUFACTURER_NAME, 'manufacturer', "utf-8")
device_info_characteristics = [manufacturer_characteristics,
                               Characteristic(CHAR_UUID_DEVICE_NAME, 'device_name', "utf-8")]

class Flags_bits( ctypes.LittleEndianStructure ):
     _fields_ = [
                 ("bit0",     c_uint8, 1 ),  # asByte & 1
                 ("bit1",     c_uint8, 1 ),  # asByte & 2
                 ("bit2",     c_uint8, 1 ),  # asByte & 4
                 ("bit3",     c_uint8, 1 ),  # asByte & 8
                 ("bit4",     c_uint8, 1 ),  # asByte & 16
                 ("bit5",     c_uint8, 1 ),  # asByte & 32
                 ("bit6",     c_uint8, 1 ),  # asByte & 64
                 ("bit7",     c_uint8, 1 ),  # asByte & 128
                ]

class Flags( ctypes.Union ):
     _anonymous_ = ("bit",)
     _fields_ = [
                 ("bit",    Flags_bits ),
                 ("asByte", c_uint8    )
                ]

class NespressoDeviceInfo:
    def __init__(self, manufacturer='', serial_nr='', model_nr='', device_name=''):
        self.manufacturer = manufacturer
        self.serial_nr = serial_nr
        self.model_nr = model_nr
        self.device_name = device_name

    def __str__(self):
        return "Manufacturer: {} Model: {} Serial: {} Device:{}".format(
            self.manufacturer, self.model_nr, self.serial_nr, self.device_name)


BYTE = Flags()
sensors_characteristics_uuid = [CHAR_UUID_STATE, CHAR_UUID_NBCAPS, CHAR_UUID_SLIDER, CHAR_UUID_WATER_HARDNESS]

sensors_characteristics_uuid_str = [str(x) for x in sensors_characteristics_uuid]


class BaseDecode:
    def __init__(self, name, format_type):
        self.name = name
        self.format_type = format_type

    def decode_data(self, raw_data):
        #val = struct.unpack(self.format_type,raw_data)
        val = raw_data
        if self.format_type == "caps_number":
            res = int.from_bytes(val,byteorder='big')
        elif self.format_type == "water_hardness":
            res = int.from_bytes(val[2:3],byteorder='big')
        elif self.format_type == "slider":
            res = binascii.hexlify(val)
            if (res) == b'00':
                res = 0
                #res = 'on'
            elif (res) == b'02':
                res = 1
                #res = 'off'
            else :
                res = "N/A"
        elif self.format_type == "state":
            BYTE0 = Flags()
            BYTE1 = Flags()
            BYTE2 = Flags()
            BYTE3 = Flags()
            
            BYTE0.asByte = val[0]
            BYTE1.asByte = val[1]
            # TODO error counter
            BYTE2.asByte = val[2]
            BYTE3.asByte = val[3]
            try:
                descaling_counter = int.from_bytes(val[6:9],byteorder='big')
            except:
                _LOGGER.debug("can't get descaling counter")
                descaling_counter = 0            
            return {"water_is_empty":BYTE0.bit0,
                    "descaling_needed":BYTE0.bit2,
                    "capsule_mechanism_jammed":BYTE0.bit4,
                    "always_1":BYTE0.bit6,
                    "water_temp_low":BYTE1.bit0,
                    "awake":BYTE1.bit1,
                    "water_engadged":BYTE1.bit2,
                    "sleeping":BYTE1.bit3,
                    "tray_sensor_during_brewing":BYTE1.bit4,
                    "tray_open_tray_sensor_full":BYTE1.bit6,
                    "capsule_engaged":BYTE1.bit7,
                    "Fault":BYTE3.bit5,
                    "descaling_counter":descaling_counter
                    }
        else:
            _LOGGER.debug("state_decoder else")
            res = val
        return {self.name:res}

sensor_decoders = {str(CHAR_UUID_STATE):BaseDecode(name="state", format_type='state'),
                   str(CHAR_UUID_NBCAPS):BaseDecode(name="caps_number", format_type='caps_number'),
                   str(CHAR_UUID_SLIDER):BaseDecode(name="slider", format_type='slider'),
                   str(CHAR_UUID_WATER_HARDNESS):BaseDecode(name="water_hardness", format_type='water_hardness'),}

class NespressoDetect:
    def __init__(self, AUTH_CODE=None, mac=None):
        self.adapter = pygatt.backends.GATTToolBackend()
        self.nespresso_devices = [] if mac is None else [mac]
        self.auth_code = AUTH_CODE
        self.sensors = []
        self.sensordata = {}
        self.keep_connected = False
        self.dev = None
    
    def set_keep_connected(self, keep_connected):
        self.keep_connected = keep_connected

    def get_info(self,tries=0):
        # Try to get some info from the discovered Nespresso devices
        self.devices = {}

        for mac in self.nespresso_devices:
            device = NespressoDeviceInfo(serial_nr=mac)
            try:
                if self.dev is None:
                    self.adapter.start(reset_on_start=False)
                    self.dev = self.adapter.connect(mac, address_type=pygatt.BLEAddressType.random, timeout=20, auto_reconnect=self.keep_connected)
                for characteristic in device_info_characteristics:
                    try:
                        data = self.dev.char_read(characteristic.uuid)
                        setattr(device, characteristic.name, data.decode(characteristic.format))
                    except (BLEError, NotConnectedError, NotificationTimeout):
                        self.dev = None
                        _LOGGER.error("Failed to read characteristics")
                if self.keep_connected is False:
                    self.dev = None
                    self.dev.disconnect()
            except (BLEError, NotConnectedError, NotificationTimeout):
                self.dev = None
                _LOGGER.debug("Failed to connect")
                time.sleep(5) # wait 5s
                if tries < 3:
                    _LOGGER.debug("Failed to get_info, 3 times")
                    self.get_info(tries+1) #retry
                else:
                    _LOGGER.error("Failed to get_info, more than 3 times")
            if self.keep_connected is False:
                self.adapter.stop()
            self.devices[mac] = device

        return self.devices

    def get_sensors(self):
        self.sensors = {}
        for mac in self.nespresso_devices:
            try:
                if self.dev is None:
                    self.adapter.start(reset_on_start=False)
                    self.dev = self.adapter.connect(mac, address_type=pygatt.BLEAddressType.random, timeout=20, auto_reconnect=self.keep_connected)
                characteristics = self.dev.discover_characteristics()
                sensor_characteristics =  []
                for characteristic in characteristics.values():
                    _LOGGER.debug(characteristic)
                    if characteristic.uuid in sensors_characteristics_uuid_str:
                        sensor_characteristics.append(characteristic)
                self.sensors[mac] = sensor_characteristics
            except (BLEError, NotConnectedError, NotificationTimeout):
                self.dev = None
                _LOGGER.error("Failed to discover sensors")
            if self.keep_connected is False:
                self.dev = None
                self.dev.disconnect()
                self.adapter.stop()

        return self.sensors
    
    def make_coffee(self, mac, temp="high", volume="lungo"):
        try:
            _LOGGER.debug("make flow a coffee")
            if self.dev is None:
                self.adapter.start(reset_on_start=False)
                self.dev = self.adapter.connect(mac, address_type=pygatt.BLEAddressType.random, timeout=20, auto_reconnect=self.keep_connected)
            self.connectnespresso(self.dev)
            try:
                characteristic = "06aa3a42-f22a-11e3-9daa-0002a5d5c51b"
                command = "0305070400000000"
                if temp == "high":
                    command += "02"
                elif temp == "low":
                    command += "01"
                else:
                    command += "00"

                if volume == "ristretto":
                    command += "00"
                elif volume == "espresso":
                    command += "01"
                elif volume == "lungo":
                    command += "02"
                elif volume == "hotwater":
                    command += "04"
                elif volume == "americano":
                    command += "05"
                else:
                    command += "00"
                self.dev.char_write(characteristic, binascii.unhexlify(command), wait_for_response=True)
            except (BLEError, NotConnectedError, NotificationTimeout):
                self.dev = None
                _LOGGER.error("Failed to write characteristic for coffee flow")
            if self.keep_connected is False:
                self.dev = None
                self.dev.disconnect()
        except (BLEError, NotConnectedError, NotificationTimeout):
            self.dev = None
            _LOGGER.error("Failed to connect for coffee flow")
        if self.keep_connected is False:
            self.adapter.stop()

    def connectnespresso(self,device,tries=0):
        try:
            #Write the auth code from android or Ios apps to the specific UUID to allow catching value from the machine
            device.char_write(CHAR_UUID_AUTH, binascii.unhexlify(self.auth_code), wait_for_response=True)
        except (BLEError, NotConnectedError, NotificationTimeout):
            self.dev = None
            _LOGGER.debug("Failed to send auth code, retrying in 5s")
            time.sleep(5) # wait 5s
            if tries < 3:
                _LOGGER.debug("Failed to send auth code, 3 times")
                self.connectnespresso(device, tries+1) #retry
            else:
                _LOGGER.error("Failed to send auth code, more than 3 times")
            
    def get_sensor_data(self):
        for mac, characteristics in self.sensors.items():
            try:
                if self.dev is None:
                    self.adapter.start(reset_on_start=False)
                    self.dev = self.adapter.connect(mac, address_type=pygatt.BLEAddressType.random, timeout=20, auto_reconnect=self.keep_connected)
                self.connectnespresso(self.dev)
                for characteristic in characteristics:
                    _LOGGER.debug("characteristic {}".format(characteristic))
                    try:
                        data = self.dev.char_read_handle("0x{:04x}".format(characteristic.handle))
                        if characteristic.uuid in sensor_decoders:
                            _LOGGER.debug("{} data {}".format(characteristic.uuid, data))
                            sensor_data = sensor_decoders[characteristic.uuid].decode_data(data)
                            #sensor_data = str(data)
                            _LOGGER.debug("{} Got sensordata {}".format(mac, sensor_data))
                            if self.sensordata.get(mac) is None:
                                self.sensordata[mac] = sensor_data
                            else:
                                self.sensordata[mac].update(sensor_data)
                    except (BLEError, NotConnectedError, NotificationTimeout):
                        self.dev = None
                        _LOGGER.error("Failed to read characteristic")

                if self.keep_connected is False:
                    self.dev = None
                    self.dev.disconnect()
            except (BLEError, NotConnectedError, NotificationTimeout):
                self.dev = None
                _LOGGER.error("Failed to connect")
            if self.keep_connected is False:
                self.adapter.stop()

        return self.sensordata


import logging
import re
import uuid
from datetime import datetime
from typing import Optional

# https://github.com/Koenkk/zigbee-herdsman-converters/blob/master/devices.js#L390
# https://slsys.io/action/devicelists.html
# All lumi models:
#   https://github.com/rytilahti/python-miio/issues/699#issuecomment-643208618
# Zigbee Model: [Manufacturer, Device Name, Device Model]
# params: [lumi res name, xiaomi prop name, hass attr name, hass domain]
DEVICES = [{
    'lumi.gateway.mgl03': ["Xiaomi", "Gateway 3", "ZNDMWG03LM"],
    'params': [
        ['8.0.2081', None, 'pairing_stop', None],
        ['8.0.2082', None, 'removed_did', None],
        ['8.0.2084', None, 'added_device', None],  # new devices added (info)
        ['8.0.2103', None, 'device_model', None],  # new device model
        ['8.0.2109', None, 'pairing_start', None],
        ['8.0.2110', None, 'discovered_mac', None],  # new device discovered
        ['8.0.2111', None, 'pair_command', None],  # add new device
        ['8.0.2155', None, 'cloud', None],  # {"cloud_link":0}
        [None, None, 'pair', 'remote'],
    ]
}, {
    # on/off, power measurement
    'lumi.plug': ["Xiaomi", "Plug", "ZNCZ02LM"],  # tested
    'lumi.plug.mitw01': ["Xiaomi", "Plug TW", "ZNCZ03LM"],
    'lumi.plug.mmeu01': ["Xiaomi", "Plug EU", "ZNCZ04LM"],
    'lumi.plug.maus01': ["Xiaomi", "Plug US", "ZNCZ12LM"],
    'lumi.ctrl_86plug': ["Aqara", "Socket", "QBCZ11LM"],
    'lumi.ctrl_86plug.aq1': ["Aqara", "Socket", "QBCZ11LM"],
    'params': [
        ['0.12.85', 'load_power', 'power', 'sensor'],
        ['0.13.85', None, 'consumption', 'sensor'],
        ['4.1.85', 'neutral_0', 'switch', 'switch'],  # or channel_0?
    ]
}, {
    'lumi.ctrl_ln1': ["Aqara", "Wall Single Switch", "QBKG11LM"],
    'lumi.ctrl_ln1.aq1': ["Aqara", "Wall Single Switch", "QBKG11LM"],
    'lumi.switch.b1nacn02': ["Aqara", "D1 Wall Single Switch", "QBKG23LM"],
    'params': [
        ['0.12.85', 'load_power', 'power', 'sensor'],
        ['0.13.85', None, 'consumption', 'sensor'],
        ['4.1.85', 'neutral_0', 'switch', 'switch'],  # or channel_0?
        ['13.1.85', None, 'action', 'sensor'],
    ]
}, {
    # dual channel on/off, power measurement
    'lumi.relay.c2acn01': ["Aqara", "Relay", "LLKZMK11LM"],  # tested
    'lumi.ctrl_ln2': ["Aqara", "Wall Double Switch", "QBKG12LM"],
    'lumi.ctrl_ln2.aq1': ["Aqara", "Wall Double Switch", "QBKG12LM"],
    'lumi.switch.b2nacn02': ["Aqara", "D1 Wall Double Switch", "QBKG24LM"],
    'params': [
        # ['0.11.85', 'load_voltage', 'power', 'sensor'],  # 0
        ['0.12.85', 'load_power', 'power', 'sensor'],
        ['0.13.85', None, 'consumption', 'sensor'],
        # ['0.14.85', None, '?', 'sensor'],  # 5.01, 6.13
        ['4.1.85', 'channel_0', 'channel 1', 'switch'],
        ['4.2.85', 'channel_1', 'channel 2', 'switch'],
        # [?, 'enable_motor_mode', 'interlock', None]
        ['13.1.85', None, 'button_1', None],
        ['13.2.85', None, 'button_2', None],
        [None, None, 'action', 'sensor'],
    ]
}, {
    # on/off
    'lumi.ctrl_neutral1': ["Aqara", "Wall Single Switch", "QBKG04LM"],
    'lumi.switch.b1lacn02': ["Aqara", "D1 Wall Single Switch", "QBKG21LM"],
    'params': [
        ['4.1.85', 'channel_0', 'switch', 'switch'],  # or neutral_0?
        ['13.1.85', None, 'action', 'sensor'],
    ]
}, {
    # dual channel on/off
    'lumi.ctrl_neutral2': ["Aqara", "Wall Double Switch", "QBKG03LM"],
    'lumi.switch.b2lacn02': ["Aqara", "D1 Wall Double Switch", "QBKG22LM"],
    'params': [
        ['4.1.85', 'channel_0', 'channel 1', 'switch'],
        ['4.2.85', 'channel_1', 'channel 2', 'switch'],
        ['13.1.85', None, 'button_1', None],
        ['13.2.85', None, 'button_2', None],
        [None, None, 'action', 'sensor'],
    ]
}, {
    # triple channel on/off, no neutral wire
    'lumi.switch.l3acn3': ["Aqara", "D1 Wall Triple Switch", "QBKG25LM"],
    'params': [
        ['4.1.85', 'channel_0', 'channel 1', 'switch'],
        ['4.2.85', 'channel_1', 'channel 2', 'switch'],
        ['4.3.85', 'channel_2', 'channel 3', 'switch'],
        ['13.1.85', None, 'button_1', None],
        ['13.2.85', None, 'button_2', None],
        ['13.3.85', None, 'button_3', None],
        [None, None, 'action', 'sensor'],
    ]
}, {
    # with neutral wire, thanks @Mantoui
    'lumi.switch.n3acn3': ["Aqara", "D1 Wall Triple Switch", "QBKG26LM"],
    'params': [
        ['0.12.85', 'load_power', 'power', 'sensor'],
        ['0.13.85', None, 'consumption', 'sensor'],
        ['4.1.85', 'channel_0', 'channel 1', 'switch'],
        ['4.2.85', 'channel_1', 'channel 2', 'switch'],
        ['4.3.85', 'channel_2', 'channel 3', 'switch'],
        ['13.1.85', None, 'button_1', None],
        ['13.2.85', None, 'button_2', None],
        ['13.3.85', None, 'button_3', None],
        [None, None, 'action', 'sensor'],
    ]
}, {
    # cube action, no retain
    'lumi.sensor_cube': ["Aqara", "Cube", "MFKZQ01LM"],
    'lumi.sensor_cube.aqgl01': ["Aqara", "Cube", "MFKZQ01LM"],  # tested
    'params': [
        ['0.2.85', None, 'duration', None],
        ['0.3.85', None, 'angle', None],
        ['13.1.85', None, 'action', 'sensor'],
        ['8.0.2001', 'battery', 'battery', 'sensor'],
    ]
}, {
    # light with brightness and color temp
    'lumi.light.aqcn02': ["Aqara", "Bulb", "ZNLDP12LM"],
    'lumi.light.cwopcn02': ["Aqara", "Opple MX650", "XDD12LM"],
    'lumi.light.cwopcn03': ["Aqara", "Opple MX480", "XDD13LM"],
    'ikea.light.led1545g12': ["IKEA", "Bulb E27 980 lm", "LED1545G12"],
    'ikea.light.led1546g12': ["IKEA", "Bulb E27 950 lm", "LED1546G12"],
    'ikea.light.led1536g5': ["IKEA", "Bulb E14 400 lm", "LED1536G5"],
    'ikea.light.led1537r6': ["IKEA", "Bulb GU10 400 lm", "LED1537R6"],
    'params': [
        ['4.1.85', 'power_status', 'light', 'light'],
        ['14.1.85', 'light_level', 'brightness', None],
        ['14.2.85', 'colour_temperature', 'color_temp', None],
    ]
}, {
    # light with brightness
    'ikea.light.led1623g12': ["IKEA", "Bulb E27 1000 lm", "LED1623G12"],
    'ikea.light.led1650r5': ["IKEA", "Bulb GU10 400 lm", "LED1650R5"],
    'ikea.light.led1649c5': ["IKEA", "Bulb E14", "LED1649C5"],  # tested
    'params': [
        ['4.1.85', 'power_status', 'light', 'light'],
        ['14.1.85', 'light_level', 'brightness', None],
    ]
}, {
    # button action, no retain
    'lumi.sensor_switch': ["Xiaomi", "Button", "WXKG01LM"],
    'lumi.sensor_switch.aq2': ["Aqara", "Button", "WXKG11LM"],
    'lumi.remote.b1acn01': ["Aqara", "Button", "WXKG11LM"],
    'lumi.sensor_switch.aq3': ["Aqara", "Shake Button", "WXKG12LM"],
    'lumi.sensor_86sw1': ["Aqara", "Single Wall Button", "WXKG03LM"],
    'lumi.remote.b186acn01': ["Aqara", "Single Wall Button", "WXKG03LM"],
    'lumi.remote.b186acn02': ["Aqara", "D1 Single Wall Button", "WXKG06LM"],
    'params': [
        ['13.1.85', None, 'button', None],
        [None, None, 'action', 'sensor'],
        ['8.0.2001', 'battery', 'battery', 'sensor'],
    ]
}, {
    # multi button action, no retain
    'lumi.sensor_86sw2': ["Aqara", "Double Wall Button", "WXKG02LM"],
    'lumi.remote.b286acn01': ["Aqara", "Double Wall Button", "WXKG02LM"],
    'lumi.sensor_86sw2.es1': ["Aqara", "Double Wall Button", "WXKG02LM"],
    'lumi.remote.b286acn02': ["Aqara", "D1 Double Wall Button", "WXKG07LM"],
    'lumi.remote.b286opcn01': ["Aqara", "Opple Two Button", "WXCJKG11LM"],
    'lumi.remote.b486opcn01': ["Aqara", "Opple Four Button", "WXCJKG12LM"],
    'lumi.remote.b686opcn01': ["Aqara", "Opple Six Button", "WXCJKG13LM"],
    'params': [
        ['13.1.85', None, 'button_1', None],
        ['13.2.85', None, 'button_2', None],
        ['13.3.85', None, 'button_3', None],
        ['13.4.85', None, 'button_4', None],
        ['13.5.85', None, 'button_both', None],
        ['13.6.85', None, 'button_5', None],
        ['13.7.85', None, 'button_6', None],
        [None, None, 'action', 'sensor'],
        ['8.0.2001', 'battery', 'battery', 'sensor'],
    ]
}, {
    # temperature and humidity sensor
    'lumi.sensor_ht': ["Xiaomi", "TH Sensor", "WSDCGQ01LM"],
    'params': [
        ['0.1.85', 'temperature', 'temperature', 'sensor'],
        ['0.2.85', 'humidity', 'humidity', 'sensor'],
        ['8.0.2001', 'battery', 'battery', 'sensor'],
    ]
}, {
    # temperature, humidity and pressure sensor
    'lumi.weather': ["Aqara", "TH Sensor", "WSDCGQ11LM"],
    'lumi.sensor_ht.agl02': ["Aqara", "TH Sensor", "WSDCGQ12LM"],
    'params': [
        ['0.1.85', 'temperature', 'temperature', 'sensor'],
        ['0.2.85', 'humidity', 'humidity', 'sensor'],
        ['0.3.85', 'pressure', 'pressure', 'sensor'],
        ['8.0.2001', 'battery', 'battery', 'sensor'],
    ]
}, {
    # door window sensor
    'lumi.sensor_magnet': ["Xiaomi", "Door Sensor", "MCCGQ01LM"],
    'lumi.sensor_magnet.aq2': ["Aqara", "Door Sensor", "MCCGQ11LM"],
    'params': [
        ['3.1.85', 'status', 'contact', 'binary_sensor'],
        ['8.0.2001', 'battery', 'battery', 'sensor'],
    ]
}, {
    # motion sensor
    'lumi.sensor_motion': ["Xiaomi", "Motion Sensor", "RTCGQ01LM"],
    'params': [
        ['3.1.85', None, 'motion', 'binary_sensor'],
        ['8.0.2001', 'battery', 'battery', 'sensor'],
    ]
}, {
    # motion sensor with illuminance
    'lumi.sensor_motion.aq2': ["Aqara", "Motion Sensor", "RTCGQ11LM"],
    'params': [
        ['0.4.85', 'illumination', 'illuminance', 'sensor'],
        ['3.1.85', None, 'motion', 'binary_sensor'],
        ['8.0.2001', 'battery', 'battery', 'sensor'],
    ]
}, {
    # water leak sensor
    'lumi.sensor_wleak.aq1': ["Aqara", "Water Leak Sensor", "SJCGQ11LM"],
    'params': [
        ['3.1.85', 'alarm', 'moisture', 'binary_sensor'],
        ['8.0.2001', 'battery', 'battery', 'sensor'],
    ]
}, {
    # vibration sensor
    'lumi.vibration.aq1': ["Aqara", "Vibration Sensor", "DJT11LM"],
    'params': [
        ['0.1.85', None, 'bed_activity', None],
        ['0.2.85', None, 'tilt_angle', None],
        ['0.3.85', None, 'vibrate_intensity', None],
        ['13.1.85', None, 'vibration', None],
        ['14.1.85', None, 'vibration_level', None],
        ['8.0.2001', 'battery', 'battery', 'sensor'],
        [None, None, 'action', 'sensor']
    ]
}, {
    'lumi.sen_ill.mgl01': ["Xiaomi", "Light Sensor", "GZCGQ01LM"],
    'params': [
        ['2.1', '2.1', 'illuminance', 'sensor'],
        ['3.1', '3.1', 'battery', 'sensor'],
    ]
}, {
    'lumi.sensor_smoke': ["Honeywell", "Smoke Sensor", "JTYJ-GD-01LM/BW"],
    'params': [
        ['0.1.85', 'density', 'smoke density', 'sensor'],
        ['13.1.85', 'alarm', 'smoke', 'binary_sensor'],
        ['8.0.2001', 'battery', 'battery', 'sensor'],
    ]
}, {
    'lumi.sensor_natgas': ["Honeywell", "Gas Sensor", "JTQJ-BF-01LM/BW"],
    'params': [
        ['0.1.85', 'density', 'gas density', 'sensor'],
        ['13.1.85', 'alarm', 'gas', 'binary_sensor'],
        ['8.0.2001', 'battery', 'battery', 'sensor'],
    ]
}, {
    'lumi.curtain': ["Aqara", "Curtain", "ZNCLDJ11LM"],
    'lumi.curtain.aq2': ["Aqara", "Roller Shade", "ZNGZDJ11LM"],
    'params': [
        ['1.1.85', 'curtain_level', 'position', None],
        ['14.2.85', None, 'motor', 'cover'],
        ['14.3.85', 'cfg_param', 'cfg_param', None],
        ['14.4.85', 'run_state', 'run_state', None],
    ]
}, {
    'lumi.curtain.hagl04': ["Aqara", "B1 Curtain", "ZNCLDJ12LM"],
    'params': [
        ['1.1.85', 'curtain_level', 'position', None],
        ['14.2.85', None, 'motor', 'cover'],
        ['14.3.85', 'cfg_param', 'cfg_param', None],
        ['14.4.85', 'run_state', 'run_state', None],
        ['8.0.2001', 'battery', 'battery', 'sensor'],
    ]
}, {  # OTHER MANUFACTURERS
    'TRADFRI bulb E27 W opal 1000lm': ["IKEA", "Bulb E27"],
    'LWB010': ["Philips", "Hue Bulb E27"],
    'FNB56-ZSC01LX1.2': ["Ali", "Dimmer"],
    'params': [
        [None, None, 'light', 'light'],
    ]
}, {
    'MS01': ["Sonoff", "Motion Sensor"],
    'params': [
        [None, None, 'motion', 'binary_sensor'],
    ]
}]

GLOBAL_PROP = {
    '8.0.2001': 'battery',
    '8.0.2002': 'reset_cnt',
    '8.0.2003': 'send_all_cnt',
    '8.0.2004': 'send_fail_cnt',
    '8.0.2005': 'send_retry_cnt',
    '8.0.2006': 'chip_temperature',
    '8.0.2007': 'lqi',
    '8.0.2008': 'voltage',
    '8.0.2009': 'pv_state',
    '8.0.2010': 'cur_state',
    '8.0.2011': 'pre_state',
    '8.0.2012': 'power_tx',
    '8.0.2013': 'CCA',
    '8.0.2014': 'protect',
    '8.0.2015': 'power',
    '8.0.2022': 'fw_ver',
    '8.0.2023': 'hw_ver',
    '8.0.2030': 'poweroff_memory',
    '8.0.2031': 'charge_protect',
    '8.0.2032': 'en_night_tip_light',
    '8.0.2034': 'load_s0',  # ctrl_dualchn
    '8.0.2035': 'load_s1',  # ctrl_dualchn
    '8.0.2036': 'parent',
    '8.0.2041': 'model',
    '8.0.2042': 'max_power',
    '8.0.2044': 'plug_detection',
    '8.0.2101': 'nl_invert',  # ctrl_86plug
    '8.0.9001': 'battery_end_of_life'
}


def get_device(zigbee_model: str) -> Optional[dict]:
    # the model has an extra tail when added
    if zigbee_model.endswith(('.v1', '.v2')):
        zigbee_model = zigbee_model[:-3]

    for device in DEVICES:
        if zigbee_model in device:
            desc = device[zigbee_model]
            return {
                # 'model': zigbee_model,
                'device_manufacturer': desc[0],
                'device_name': desc[0] + ' ' + desc[1],
                'device_model': zigbee_model + ' ' + desc[2]
                if len(desc) > 2 else zigbee_model,
                'params': device['params']
            }

    return None



import asyncio
import functools
import glob
import os

import cv2
import numpy as np
from tflite_runtime.interpreter import Interpreter, load_delegate

from .. import fhem, utils
from threading import Thread
import time
import concurrent

class object_detection:

    def __init__(self, logger):
        self.logger = logger
        self.loop = asyncio.get_event_loop()
        self._cwd_path = os.getcwd()
        obj_det_mod_dir = os.path.dirname(os.path.abspath(__file__))
        self._labels_path = os.path.join(obj_det_mod_dir, "labelmap.txt")
        self._graph_path = os.path.join(obj_det_mod_dir, "detect.tflite")
        self._detection_task = None
        self._prev_objects = None
        self._stop_detection = False
        self._attr_detection_interval = 2
        self._attr_detection_threshold = 0.6
        self._attr_list = {
            "detection_interval": { "default": 2, "format": "float" },
            "detection_threshold": { "default": 0.6, "format": "float" }
        }
        return

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        self.hash = hash
        if len(args) < 5:
            return "Usage: define obj_detection PythonModule object_detection <image/stream> <LOCATION>"
        if args[3] == "image":
            if args[4][0] == "/":
                self._source_uri = args[4]
            else:
                self._source_uri = os.path.join(self._cwd_path, args[4])
        else:
            self._source_uri = args[4]
        self._source_type = args[3]

        await utils.handle_define_attr(self._attr_list, self, hash)

        self.logger.debug(f"Source URI: {self._source_uri}")
        await fhem.readingsSingleUpdate(self.hash, "state", "stopped", 1)
        return ""
    
    async def Attr(self, hash, args, argsh):
        return await utils.handle_attr(self._attr_list, self, hash, args, argsh)

    def run_stream_object_detection(self):
        # Load the label map
        with open(self._labels_path, 'r') as f:
            labels = [line.strip().replace(" ", "_") for line in f.readlines()]

        # Have to do a weird fix for label map if using the COCO "starter model" from
        # https://www.tensorflow.org/lite/models/object_detection/overview
        # First label is '???', which has to be removed.
        if labels[0] == '???':
            del(labels[0])
        interpreter = Interpreter(model_path=self._graph_path)
        interpreter.allocate_tensors()

        # Get model details
        input_details = interpreter.get_input_details()
        output_details = interpreter.get_output_details()
        height = input_details[0]['shape'][1]
        width = input_details[0]['shape'][2]
        floating_model = (input_details[0]['dtype'] == np.float32)
        input_mean = 127.5
        input_std = 127.5

        # Initialize frame rate calculation
        frame_rate_calc = 1
        freq = cv2.getTickFrequency()

        # resolution
        imW = 1280
        imH = 720

        # Initialize video stream
        videostream = VideoStream(self._source_uri, resolution=(imW,imH),framerate=30,loop=self.loop).start()
        time.sleep(1)

        #for frame1 in camera.capture_continuous(rawCapture, format="bgr",use_video_port=True):
        while True:

            try:
                # Start timer (for calculating frame rate)
                #t1 = cv2.getTickCount()

                # Grab frame from video stream
                frame1 = videostream.read()
                if frame1 is None:
                    continue

                # Acquire frame and resize to expected shape [1xHxWx3]
                frame = frame1.copy()
                frame_rgb = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                frame_resized = cv2.resize(frame_rgb, (width, height))
                input_data = np.expand_dims(frame_resized, axis=0)

                # Normalize pixel values if using a floating model (i.e. if model is non-quantized)
                if floating_model:
                    input_data = (np.float32(input_data) - input_mean) / input_std

                # Perform the actual detection by running the model with the image as input
                interpreter.set_tensor(input_details[0]['index'],input_data)
                interpreter.invoke()

                # Retrieve detection results
                boxes = interpreter.get_tensor(output_details[0]['index'])[0] # Bounding box coordinates of detected objects
                classes = interpreter.get_tensor(output_details[1]['index'])[0] # Class index of detected objects
                scores = interpreter.get_tensor(output_details[2]['index'])[0] # Confidence of detected objects
                #num = interpreter.get_tensor(output_details[3]['index'])[0]  # Total number of detected objects (inaccurate and not needed)

                detected_objects = []
                # Loop over all detections and draw detection box if confidence is above minimum threshold
                for i in range(len(scores)):
                    if ((scores[i] > self._attr_detection_threshold) and (scores[i] <= 1.0)):

                        # Get bounding box coordinates and draw box
                        # Interpreter can return coordinates that are outside of image dimensions, need to force them to be within image using max() and min()
                        # ymin = int(max(1,(boxes[i][0] * imH)))
                        # xmin = int(max(1,(boxes[i][1] * imW)))
                        # ymax = int(min(imH,(boxes[i][2] * imH)))
                        # xmax = int(min(imW,(boxes[i][3] * imW)))
                        
                        # cv2.rectangle(frame, (xmin,ymin), (xmax,ymax), (10, 255, 0), 2)

                        # Draw label
                        object_name = labels[int(classes[i])] # Look up object name from "labels" array using class index
                        label = '%s: %d%%' % (object_name, int(scores[i]*100)) # Example: 'person: 72%'
                        # labelSize, baseLine = cv2.getTextSize(label, cv2.FONT_HERSHEY_SIMPLEX, 0.7, 2) # Get font size
                        # label_ymin = max(ymin, labelSize[1] + 10) # Make sure not to draw label too close to top of window
                        # cv2.rectangle(frame, (xmin, label_ymin-labelSize[1]-10), (xmin+labelSize[0], label_ymin+baseLine-10), (255, 255, 255), cv2.FILLED) # Draw white box to put label text in
                        # cv2.putText(frame, label, (xmin, label_ymin-7), cv2.FONT_HERSHEY_SIMPLEX, 0.7, (0, 0, 0), 2) # Draw label text

                        detected_objects.append({"object": object_name, "score": int(scores[i]*100)})

                asyncio.run_coroutine_threadsafe(self.update_readings(detected_objects), self.loop).result()

                if self._stop_detection:
                    asyncio.run_coroutine_threadsafe(fhem.readingsSingleUpdate(self.hash, "state", "stopped", 1), self.loop).result()
                    videostream.stop()
                    return

                # Draw framerate in corner of frame
                #cv2.putText(frame,'FPS: {0:.2f}'.format(frame_rate_calc),(30,50),cv2.FONT_HERSHEY_SIMPLEX,1,(255,255,0),2,cv2.LINE_AA)

                # All the results have been drawn on the frame, so it's time to display it.
                #cv2.imshow('Object detector', frame)

                # Calculate framerate
                # t2 = cv2.getTickCount()
                # time1 = (t2-t1)/freq
                # frame_rate_calc= 1/time1
                time.sleep(self._attr_detection_interval)
            except asyncio.CancelledError:
                videostream.stop()
                return

    def run_image_object_detection(self):
        detected_objects = []
        # Load the label map
        with open(self._labels_path, 'r') as f:
            labels = [line.strip() for line in f.readlines()]

        # Have to do a weird fix for label map if using the COCO "starter model" from
        # https://www.tensorflow.org/lite/models/object_detection/overview
        # First label is '???', which has to be removed.
        if labels[0] == '???':
            del(labels[0])
        interpreter = Interpreter(model_path=self._graph_path)
        interpreter.allocate_tensors()

        # Get model details
        input_details = interpreter.get_input_details()
        output_details = interpreter.get_output_details()
        height = input_details[0]['shape'][1]
        width = input_details[0]['shape'][2]
        floating_model = (input_details[0]['dtype'] == np.float32)
        input_mean = 127.5
        input_std = 127.5

        # Load image and resize to expected shape [1xHxWx3]
        image_path = self._source_uri
        image = cv2.imread(image_path)
        image_rgb = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        imH, imW, _ = image.shape 
        image_resized = cv2.resize(image_rgb, (width, height))
        input_data = np.expand_dims(image_resized, axis=0)

        # Normalize pixel values if using a floating model (i.e. if model is non-quantized)
        if floating_model:
            input_data = (np.float32(input_data) - input_mean) / input_std

        # Perform the actual detection by running the model with the image as input
        interpreter.set_tensor(input_details[0]['index'],input_data)
        interpreter.invoke()

        # Retrieve detection results
        boxes = interpreter.get_tensor(output_details[0]['index'])[0] # Bounding box coordinates of detected objects
        classes = interpreter.get_tensor(output_details[1]['index'])[0] # Class index of detected objects
        scores = interpreter.get_tensor(output_details[2]['index'])[0] # Confidence of detected objects
        #num = interpreter.get_tensor(output_details[3]['index'])[0]  # Total number of detected objects (inaccurate and not needed)

        # Loop over all detections and draw detection box if confidence is above minimum threshold
        for i in range(len(scores)):
            if ((scores[i] > self._attr_detection_threshold) and (scores[i] <= 1.0)):
    
                # Get bounding box coordinates and draw box
                # Interpreter can return coordinates that are outside of image dimensions, need to force them to be within image using max() and min()
                ymin = int(max(1,(boxes[i][0] * imH)))
                xmin = int(max(1,(boxes[i][1] * imW)))
                ymax = int(min(imH,(boxes[i][2] * imH)))
                xmax = int(min(imW,(boxes[i][3] * imW)))
                
                cv2.rectangle(image, (xmin,ymin), (xmax,ymax), (10, 255, 0), 2)
    
                object_name = labels[int(classes[i])] # Look up object name from "labels" array using class index
                detected_objects.append({"object": object_name, "score": int(scores[i]*100)})

                # Draw label
                label = '%s: %d%%' % (object_name, int(scores[i]*100)) # Example: 'person: 72%'
                labelSize, baseLine = cv2.getTextSize(label, cv2.FONT_HERSHEY_SIMPLEX, 0.7, 2) # Get font size
                label_ymin = max(ymin, labelSize[1] + 10) # Make sure not to draw label too close to top of window
                cv2.rectangle(image, (xmin, label_ymin-labelSize[1]-10), (xmin+labelSize[0], label_ymin+baseLine-10), (255, 255, 255), cv2.FILLED) # Draw white box to put label text in
                cv2.putText(image, label, (xmin, label_ymin-7), cv2.FONT_HERSHEY_SIMPLEX, 0.7, (0, 0, 0), 2) # Draw label text

                # output file with object rectangle and text
                #cv2.imwrite(self._output_file, image)
        return detected_objects

    # FHEM FUNCTION
    async def Undefine(self, hash):
        if self._detection_task:
            self._detection_task.cancel()
        return

    # FHEM FUNCTION
    async def Set(self, hash, args, argsh):
        set_list_conf = {
           "start": { },
           "detect_once": { },
           "stop": { }
        }
        return await utils.handle_set(set_list_conf, self, hash, args, argsh)

    async def set_start(self, hash):
        self._stop_detection = False
        await self.start_detection()

    async def set_detect_once(self, hash):
        self._stop_detection = True
        await self.start_detection()

    async def start_detection(self):
        await fhem.readingsSingleUpdate(self.hash, "state", "running", 1)
        if self._source_type == "image":
            asyncio.create_task(self.image_detect_objects_loop())
        else:
            self._detection_task = utils.run_blocking_task(functools.partial(self.run_stream_object_detection))
        return ""

    async def set_stop(self, hash):
        self._stop_detection = True

    async def image_detect_objects_loop(self):
        while True:
            await self.image_detect_objects(self)
            if self._stop_detection:
                break
            await asyncio.sleep(self._attr_detection_interval)
        await fhem.readingsSingleUpdate(self.hash, "state", "stopped", 1)

    async def image_detect_objects(self):
        detected_objects = await utils.run_blocking(functools.partial(self.run_image_object_detection))
        await self.update_readings(detected_objects)

    async def update_readings(self, detected_objects):
        try:
            all_objects = {}
            curr_objects = {}
            if self._prev_objects is None:
                await fhem.CommandDeleteReading(self.hash, self.hash["NAME"] + " object_.*")

            await fhem.readingsBeginUpdate(self.hash)
            for obj in detected_objects:
                obj_name = obj['object']
                obj_score = obj['score']
                if obj_name not in all_objects:
                    all_objects[obj_name] = 0
                all_objects[obj_name] += 1
                await fhem.readingsBulkUpdateIfChanged(self.hash, "object_" + obj_name, obj_score)
            for obj_name in all_objects:
                await fhem.readingsBulkUpdateIfChanged(self.hash, "object_count_" + obj_name, all_objects[obj_name])
                curr_objects[obj_name] = all_objects[obj_name]
            
            # set readings to 0 for objects which were not found
            if self._prev_objects is None:
                set_0_readings = {}
            else:
                set_0_readings = set(self._prev_objects) - set(curr_objects)
            for set_0_reading in set_0_readings:
                await fhem.readingsBulkUpdateIfChanged(self.hash, "object_" + set_0_reading, 0)
                await fhem.readingsBulkUpdateIfChanged(self.hash, "object_count_" + set_0_reading, 0)
            await fhem.readingsBulkUpdateIfChanged(self.hash, "objects_detected", ",".join(set(all_objects)))
            await fhem.readingsEndUpdate(self.hash, 1)

            self._prev_objects = curr_objects
        except:
            self.logger.exception("Failed to update readings")

# Define VideoStream class to handle streaming of video from webcam in separate processing thread
# Source - Adrian Rosebrock, PyImageSearch: https://www.pyimagesearch.com/2015/12/28/increasing-raspberry-pi-fps-with-python-and-opencv/
class VideoStream:
    """Camera object that controls video streaming"""
    def __init__(self,source_uri,resolution=(640,480),framerate=30,loop=None):
        self.loop = loop
        # Initialize the PiCamera and the camera image stream
        self.stream = cv2.VideoCapture(source_uri)
        ret = self.stream.set(cv2.CAP_PROP_FOURCC, cv2.VideoWriter_fourcc(*'MJPG'))
        ret = self.stream.set(cv2.CAP_PROP_FPS, framerate)
        ret = self.stream.set(3,resolution[0])
        ret = self.stream.set(4,resolution[1])
            
        # Read first frame from the stream
        (self.grabbed, self.frame) = self.stream.read()

	    # Variable to control when the camera is stopped
        self.stopped = False

    def start(self):
	    # Start the thread that reads frames from the video stream
        self._task = self.loop.create_task(self.update_task())
        return self
    
    async def update_task(self):
        with concurrent.futures.ThreadPoolExecutor() as pool:
            return await self.loop.run_in_executor(
                pool, functools.partial(self.update))

    def update(self):
        # Keep looping indefinitely until the thread is stopped
        while True:
            try:
                # If the camera is stopped, stop the thread
                if self.stopped:
                    # Close camera resources
                    self.stream.release()
                    return

                # Otherwise, grab the next frame from the stream
                (self.grabbed, self.frame) = self.stream.read()
            except asyncio.CancelledError:
                self.stream.release()
                return

    def read(self):
	    # Return the most recent frame
        return self.frame

    def stop(self):
	    # Indicate that the camera and thread should be stopped
        self.stopped = True